import logging
import asyncio
import os
from typing import Callable

import telegram
from telegram.constants import ChatType, ChatMemberStatus
from telegram.ext import Application, ContextTypes, CommandHandler, MessageHandler, filters


from .message.telegram import TelegramReflectionMessage
from ..caching.base import MediaCache, UnsupportedMediaException
from ..enums import Platform, Messages

ImageCache = None
try:
    from ..caching import ImageCache
except ImportError:
    pass


class TelegramReflectionAPI:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

        self.platform = Platform.TELEGRAM

        self.token = os.getenv("TELEGRAM_BOT_TOKEN")
        if self.token is None:
            raise RuntimeError("TELEGRAM_BOT_TOKEN environment variable is not set, unable to initialize telegram bot.")

        application = Application.builder() \
            .token(self.token) \
            .read_timeout(30) \
            .write_timeout(30) \
            .build()
        self.application = application

    def run(self):
        self.application.run_polling()

    def register_on_message_event(self, function: Callable):
        async def on_message(update: telegram.Update, context: ContextTypes.DEFAULT_TYPE):
            message = update.message
            assert message is not None
            message = TelegramReflectionMessage.from_origin(message)
            if message is not None:
                if asyncio.iscoroutinefunction(function):
                    return await function(message)
                else:
                    return function(message)

        self.application.add_handler(MessageHandler(
            filters.TEXT | filters.VIDEO | filters.AUDIO | filters.Document.ALL,
            callback=on_message
        ))

    def register_slash_command(self, name: str, function: Callable, description: str | None = None):
        async def slash_command(update: telegram.Update, context: ContextTypes.DEFAULT_TYPE):
            message = update.message
            assert message is not None
            message = TelegramReflectionMessage.from_origin(message)
            if message is not None:
                if asyncio.iscoroutinefunction(function):
                    return await function(message)
                else:
                    return function(message)
        self.application.add_handler(CommandHandler(name, slash_command))

    def get_guild_info(self, message: TelegramReflectionMessage) -> dict:
        return dict()

    def get_thread_info(self, message: TelegramReflectionMessage) -> dict:
        return dict()

    def can_read_history(self, channel) -> bool:
        return True

    def is_message_from_the_bot(self, message: TelegramReflectionMessage) -> bool:
        origin: telegram.Message = message.origin
        assert origin.from_user is not None, "from_user is None"
        return origin.get_bot().id == origin.from_user.id

    def is_bot_mentioned(self, message: TelegramReflectionMessage):
        return f"@{self.application.bot.username}" in message.content

    async def is_dm_or_admin(self, message: TelegramReflectionMessage) -> bool:
        if message.environment.chat_type == ChatType.PRIVATE:
            return True
        # Check if user has admin permissions to use the bot commands
        member = await self.application.bot.get_chat_member(message.environment.chat_id, message.author.id)
        return member.status in (ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER)

    async def send_file(self, chat_id: int, filepath: str, filename: str, caption: str | None = None):
        chat = await self.application.bot.get_chat(chat_id)
        for i in range(3):
            try:
                with open(filepath, "rb") as f:
                    await chat.send_document(
                        document=f,
                        filename=filename,
                        caption=caption,
                        read_timeout=30,
                        write_timeout=600,
                    )
            except telegram.error.TimedOut as e:
                self.logger.error(f"Timed out while sending file: {e}, retrying {i+1}/{3}")
            except telegram.error.BadRequest as e:
                self.logger.exception(f"BadRequest error while sending file: {e}")
                break
            except Exception as e:
                self.logger.exception(f"Unexpected error while sending file: {e}")
            else:
                break

    async def get_user_avatar(self, user_id: int) -> MediaCache | None:
        try:
            file = await self.application.bot.get_user_profile_photos(user_id)
            if file.photos:
                photo = file.photos[0][-1]  # Get the highest resolution photo
                assert ImageCache
                image_bytes = await (await photo.get_file()).download_as_bytearray()
                return ImageCache(bytes(image_bytes))
        except telegram.error.BadRequest as e:
            self.logger.error(f"Failed to fetch user avatar: {e}")
            return None
        except Exception as e:
            self.logger.exception(f"Unexpected error while fetching user avatar: {e}")
            return None

    async def fetch_channel_history(self, channel_id: int, n: int = 10):
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("fetch_channel_history", self.platform))



import dataclasses

@dataclasses.dataclass(frozen=True)
class OpenAIAuthConfig:
    base_url: str
    api_key: str

@dataclasses.dataclass(frozen=True)
class OpenAILanguageModelConfig:
    id: str
    max_context: int
    
@dataclasses.dataclass(frozen=True)
class OpenAIEmbeddingModelConfig:
    id: str
    max_context: int
    dimension: int
    min_chunk_size: int
    max_chunk_size: int
    


from enum import StrEnum

from .messages import Messages


class Platform(StrEnum):
    DISCORD = "discord"
    TELEGRAM = "telegram"


class ChatRole(StrEnum):
    SYSTEM = "system"
    ASSISTANT = "assistant"
    USER = "user"
    TOOL = "tool"


class ChatType(StrEnum):
    PRIVATE = "private"
    GROUP = "group"
    TEXT = "text"
    UNKNOWN = "unknown"



from typing import Any, Callable, Coroutine, Optional, Iterator, Generator, AsyncIterator, AsyncGenerator

AsyncFunction = Callable[..., Coroutine[Any, Any, Any]]
AsyncPredicate = Callable[..., Coroutine[Any, Any, bool]]



import logging

from .base import UnsupportedMediaException, MediaCache

logger = logging.getLogger("pixi.caching")

SUPPORTS_MEDIA_CACHING = False
try:
    import av
    import av.logging
    av.logging.set_level(av.logging.VERBOSE)
    SUPPORTS_MEDIA_CACHING = True
except ImportError:
    logger.warning("please install `av` to use media caching features")
else:
    from .audiocache import AudioCache
    from .imagecache import ImageCache

try:
    from .embedding import EmbedingCache
except ImportError:
    logger.warning("please install `numpy` to use embedding vector caching features")



import os
import hashlib
from base64 import b64encode
from dataclasses import dataclass
from typing import Optional

from ..utils import exists


class UnsupportedMediaException(Exception):
    """Exception raised when an unsupported media type or format is encountered."""

    def __init__(self, message="Unsupported media type or format."):
        super().__init__(message)


@dataclass
class CompressedMedia:
    mime_type: str
    bytes: bytes
    format: Optional[str] = None
    metadata: Optional[dict] = None

    def to_base64(self):
        return b64encode(self.bytes).decode("utf-8")

    def to_data_url(self):
        return "data:image/jpeg;base64," + self.to_base64()


class MediaCache:
    """
    Handles efficient caching of media by generating compressed versions and storing them on disk.
    - If initialized with data_bytes, it creates a hash, optimizes the image, and saves it to the cache.
    - If initialized with a hash, it loads the cached media if available.
    Provides methods to retrieve the media as bytes, or a data URL, and to check cache existence.
    """

    def __init__(self, cache_dir: str, format: str, mime_type: str, data_bytes: Optional[bytes] = None, hash_value: Optional[str] = None):
        assert isinstance(cache_dir, str), f"expected `cache_dir` to be of type str but got {cache_dir}."
        assert isinstance(format, str), f"expected `format` to be of type str but got {format}."
        assert isinstance(mime_type, str), f"expected `mime_type` to be of type str but got {mime_type}."
        assert exists(data_bytes) ^ exists(hash_value), "Only `image_bytes` OR `hash_value` must be provided."

        self.cache_dir = cache_dir
        self.hash = None
        self.media = None
        self.format = format
        self.mime_type = mime_type

        if exists(data_bytes):
            assert isinstance(data_bytes, bytes), "image_bytes must be of type bytes."
            self.hash = self.compute_hash(data_bytes)
            if self.exists():
                self.load_cache()
            else:
                self.media = self.compress(data_bytes)

                os.makedirs(self.cache_dir, exist_ok=True)

                path = self.cache_path
                if path and not os.path.isfile(path):
                    with open(path, "wb") as f:
                        f.write(self.media.bytes)

        if exists(hash_value):
            assert isinstance(hash_value, str), "hash_value must be of type str."
            self.hash = hash_value
            self.load_cache()

    @staticmethod
    def compute_hash(data_bytes: bytes) -> str:
        # hash only the first 16 megabytes for performance
        return hashlib.sha256(data_bytes[:2**24]).hexdigest()

    @property
    def cache_path(self) -> Optional[str]:
        if self.hash:
            return os.path.join(self.cache_dir, f"{self.hash}.{self.format}")
        return None

    def compress(self, data_bytes: bytes) -> CompressedMedia:
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, data: dict):
        if not (hash_value := data.get("hash")):
            raise ValueError("Hash value is required to load ImageCache instance.")
        return cls(hash_value=hash_value)  # type: ignore

    def to_dict(self) -> dict:
        return dict(hash=self.hash)

    def exists(self) -> bool:
        path = self.cache_path
        return exists(path) and os.path.isfile(path)  # type: ignore

    def __ensure_initialized(self):
        if self.media is None:
            raise RuntimeError("attemted to access uninitialized cache")

    def load_cache(self):
        if self.media is None:
            path = self.cache_path
            if path and os.path.exists(path):
                with open(path, "rb") as f:
                    self.media = CompressedMedia(
                        mime_type=self.mime_type,
                        bytes=f.read(),
                        format=self.format,
                    )
            else:
                raise FileNotFoundError("Media not found in cache.")
        return self.media

    def get_bytes(self) -> Optional[bytes]:
        if self.media is None:
            self.load_cache()
        assert self.media
        return self.media.bytes

    def get_mime_type(self) -> str:
        return self.mime_type

    def get_format(self) -> str:
        return self.format

    def to_data_url(self) -> str:
        self.__ensure_initialized()
        assert self.media
        return self.media.to_data_url()

    def to_base64(self) -> str:
        self.__ensure_initialized()
        assert self.media
        return self.media.to_base64()



import os
import hashlib
from typing import Optional

import numpy as np

from ..utils import PixiPaths


class EmbedingCache:
    """
    Automatically caches the embedding vectors if initialized with `vec` and automatically
    loads the cahced embedding vectors if not initialized with `vec`, raises OSError if the
    cache file does not exist or cannot be read.
    """

    def __init__(self, text: str, dim: int, vec: Optional[np.ndarray] = None):
        self.text = text
        self.dim = dim
        self.vec = vec

        cache_dir = PixiPaths.cache() / "embeddings"

        if self.vec is None:
            # loads the cached vector
            self.load()
        else:
            # caches the vector
            os.makedirs(cache_dir, exist_ok=True)
            self.save()

    @property
    def hash_hexdigest(self):
        return hashlib.sha256(str(dict(text=self.text, dim=self.dim)).encode("utf-8")).hexdigest()

    @staticmethod
    def get_save_path(hash: str):
        return str(PixiPaths.cache() / "embeddings" / f"{hash}.npz")

    def save(self):
        assert self.vec is not None
        return np.savez_compressed(self.get_save_path(self.hash_hexdigest), vec=self.vec)

    def load(self):
        data = np.load(self.get_save_path(self.hash_hexdigest))
        self.vec = data["vec"]



__author__ = "amiralimollaei"

import os
from pathlib import Path

from .config import OpenAIAuthConfig, OpenAIEmbeddingModelConfig, OpenAILanguageModelConfig, PixiFeatures, IdFilter
from .enums import Platform


def run(
    platform: Platform,
    pixi_directory: str | Path,
    *,
    auth: OpenAIAuthConfig,
    model: OpenAILanguageModelConfig,
    helper_model: OpenAILanguageModelConfig | None,
    embedding_model: OpenAIEmbeddingModelConfig | None,
    features: PixiFeatures = PixiFeatures.empty(),
    environment_filter: IdFilter = IdFilter.allow(),
    database_names: list[str] | None = None,
):
    from .utils import PixiPaths, copy_default_resources
    PixiPaths.set_root(pixi_directory)
    copy_default_resources()
    
    from .client import PixiClient
    
    PixiClient(
        platform=platform,
        auth=auth,
        model=model,
        helper_model=helper_model,
        embedding_model=embedding_model,
        features=features,
        environment_filter=environment_filter,
        database_names=database_names,
    ).run()


def main():
    import logging

    import argparse

    from .utils import load_dotenv, Ansi

    # injecting colors into the default logger

    COLORS = {
        logging.DEBUG: (Ansi.BLUE, Ansi.BEIGE),
        logging.INFO: (Ansi.WHITE, Ansi.WHITE2),
        logging.WARNING: (Ansi.YELLOW, Ansi.YELLOW2),
        logging.ERROR: (Ansi.RED, Ansi.RED2),
        logging.CRITICAL: (Ansi.BOLD + Ansi.RED, Ansi.BOLD + Ansi.RED2)
    }

    orig_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record = orig_factory(*args, **kwargs)
        level_color, message_color = COLORS.get(record.levelno, (Ansi.GREY, Ansi.WHITE))
        record.level_color = level_color
        record.message_color = message_color
        return record
    logging.setLogRecordFactory(record_factory)

    logging.basicConfig(
        format=f"{Ansi.GREY}[{Ansi.BLUE}%(asctime)s{Ansi.GREY}] {Ansi.GREY}[%(level_color)s%(levelname)s / %(name)s{Ansi.GREY}] %(message_color)s%(message)s{Ansi.END}",
        level=logging.INFO,
        force=True,
    )

    # https://github.com/langchain-ai/langchain/issues/14065#issuecomment-1834571761
    # Get the logger for 'httpx'
    httpx_logger = logging.getLogger("httpx")
    # Set the logging level to WARNING to ignore INFO and DEBUG logs
    httpx_logger.setLevel(logging.WARNING)

    # load environment variables
    load_dotenv()

    parser = argparse.ArgumentParser(description="Run the Pixi bot, a multi-platform AI chatbot.")
    parser.add_argument(
        "--platform", "-p",
        type=str,
        choices=[p.name.lower() for p in Platform],
        required=True,
        help="Platform to run the bot on."
    )
    parser.add_argument(
        "--pixi-directory", "-pd",
        type=str,
        choices=[p.name.lower() for p in Platform],
        default="~/.pixi/",
        help="The root directory for configuration files, addons, userdata, assets and cache, defaults to \"~/.pixi/\""
    )
    parser.add_argument(
        "--log-level", "-l",
        type=str,
        choices=["debug", "info", "warning", "error", "critical"],
        default="info",
        help="Set the logging level."
    )
    parser.add_argument(
        "--api-url", "-a",
        type=str,
        default="https://api.openai.com/v1",
        help="OpenAI Compatible API URL to use for the bot"
    )
    parser.add_argument(
        "--auth",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="whether or not to authorize to the API backends"
    )

    # model arguments
    parser.add_argument(
        "--model", "-m",
        type=str,
        required=True,
        help="Language Model to use for the main chatbot bot"
    )
    parser.add_argument(
        "--model-max-context", "-ctx",
        type=int,
        default=16192,
        help="Maximum model context size (in tokens), pixi tries to apporiximately stay within this context size, Default is '16192`."
    )

    # helper model arguments
    parser.add_argument(
        "--helper-model", "-hm",
        type=str,
        help="Language Model to use for agentic tools"
    )
    parser.add_argument(
        "--helper-model-max-context", "-hctx",
        type=int,
        default=16192,
        help="Maximum helper model context size (in tokens), pixi tries to apporiximately stay within this context size, Default is '16192`."
    )

    # embedding model arguments
    parser.add_argument(
        "--embedding-model", "-em",
        type=str,
        help="Embedding Model to use for embedding tools"
    )
    parser.add_argument(
        "--embedding-model-max-context", "-ectx",
        type=int,
        default=16192,
        help="Maximum embedding model context size (in tokens), pixi tries to apporiximately stay within this context size, Default is '16192`."
    )
    parser.add_argument(
        "--embedding-model-dimension", "-ed",
        type=int,
        default=768,
        help="Dimention to use for the embedding model, Default is '768`."
    )
    parser.add_argument(
        "--embedding-model-min-size", "-emin",
        type=int,
        default=256,
        help="Minimum chunk size to use for the embedding chunk tokenizer, Default is '256`."
    )
    parser.add_argument(
        "--embedding-model-max-size", "-emax",
        type=int,
        default=1024,
        help="Maximum chunk size to use for the embedding chunk tokenizer, Default is '1024`."
    )

    # feature arguments
    parser.add_argument(
        "--tool-calling",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="allows pixi to use built-in and/or plugin tools, tool calling can only be used if the model supports them"
    )
    parser.add_argument(
        "--tool-logging",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="verbose logging for tool calls (enabled by default when running with logging level DEBUG)"
    )
    parser.add_argument(
        "--wiki-search",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="allows pixi to search any mediawiki compatible Wiki"
    )
    parser.add_argument(
        "--gif-search",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="allows pixi to search for gifs online, and send them in chat"
    )
    parser.add_argument(
        "--image-support",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="allows pixi to download and process image files"
    )
    parser.add_argument(
        "--audio-support",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="allows pixi to download and process audio files"
    )

    # environment filter arguments
    parser.add_argument(
        "--environment-whitelist",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="whether or not the ids passed to environment ids are whitelisted or blacklisted"
    )
    parser.add_argument(
        "--environment-ids",
        type=str,
        nargs="+",
        default=None,
        help="add the id of the environment that the bot is or is not allowed to respond in (space-separated). If not provided, the bot will respond everywhere."
    )

    # database arguments
    parser.add_argument(
        "--database-names", "-d",
        type=str,
        nargs="+",
        default=None,
        help="add the name of databases to use (space-separated)."
    )
    args = parser.parse_args()

    # Set logging level
    logging.root.setLevel(args.log_level.upper())

    api_key = "NO_AUTH"
    if args.auth:
        api_key = os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise RuntimeError("OPENAI_API_KEY environment variable is not set.")

    auth = OpenAIAuthConfig(
        base_url=os.getenv("OPENAI_BASE_URL", args.api_url),
        api_key=api_key
    )

    model = OpenAILanguageModelConfig(
        id=args.model,
        max_context=args.model_max_context,
    )

    helper_model = OpenAILanguageModelConfig(
        id=args.helper_model,
        max_context=args.helper_model_max_context,
    ) if args.helper_model else None

    embedding_model = OpenAIEmbeddingModelConfig(
        id=args.embedding_model,
        max_context=args.embedding_model_max_context,
        dimension=args.embedding_model_dimension,
        min_chunk_size=args.embedding_model_min_size,
        max_chunk_size=args.embedding_model_max_size,
    ) if args.embedding_model else None

    features = PixiFeatures(0)
    if args.tool_calling:
        features |= PixiFeatures.EnableToolCalling
    if args.tool_logging:
        features |= PixiFeatures.EnableToolLogging
    if args.wiki_search:
        features |= PixiFeatures.EnableWikiSearch
    if args.gif_search:
        features |= PixiFeatures.EnableGIFSearch
    if args.image_support:
        features |= PixiFeatures.EnableImageSupport
    if args.audio_support:
        features |= PixiFeatures.EnableAudioSupport

    environment_filter = IdFilter.allow()
    if args.environment_ids:
        environment_filter = IdFilter.whitelist(
            args.environment_ids) if args.environment_whitelist else IdFilter.blacklist(args.environment_ids)

    return run(
        Platform[args.platform.upper()],
        pixi_directory=args.pixi_directory,
        auth=auth,
        model=model,
        helper_model=helper_model,
        embedding_model=embedding_model,
        features=features,
        environment_filter=environment_filter,
        database_names=args.database_names,
    )



from typing import Optional

from .api import APIBase


BASE_URL = "https://g.tenor.com/v1/"


class AsyncTenorAPI(APIBase):

    def __init__(self, api_key: Optional[str] = None):
        super().__init__(base=BASE_URL, api_key=api_key, api_key_env_var="TENOR_API_KEY", api_key_key="key")

    async def search(
        self,
        q: str,
        limit: Optional[int] = None,
        locale: Optional[str] = None,
        media_filter: Optional[str] = "minimal",
    ):
        return await self.request("search", data=dict(
            q=q,
            limit=limit,
            locale=locale,
            media_filter=media_filter,
        ))



import json
import os
from typing import Optional

import httpx


class APIBase:
    """
    Base class for asynchronous API clients using httpx.
    This class provides a method to make GET requests to an API endpoint with optional parameters and an API key.
    It requires a base URL for the API and optionally an API key that can be provided either directly or through an environment variable.

    Attributes:
        base (str): The base URL of the API.
        api_key_env_var (str): The name of the environment variable containing the API key.
        api_key (Optional[str]): The API key to use for requests, if not provided through the environment variable.

    Raises:
        KeyError: If the base URL is None or if the API key is not provided and the environment variable is not set.
    """

    def __init__(self, base: str | None , api_key_env_var: Optional[str] = None, api_key: Optional[str] = None, api_key_key: str = "api_key"):
        assert base, f"base url is not valid, got \"{base}\""
        self.base = base

        if api_key_env_var:
            api_key = api_key or os.getenv(api_key_env_var)
            if api_key is None:
                raise KeyError(f"{api_key_env_var} must be set in environment variables or passed as an argument.")

        self.api_key = api_key
        self.api_key_key = api_key_key

        self.client = httpx.AsyncClient(trust_env=False)

    async def request(self, url: str, data: Optional[dict] = None) -> str | dict | list:
        api_key_data = {self.api_key_key: self.api_key} if self.api_key else dict()
        if data is None:
            data = api_key_data
        else:
            data = data | api_key_data
        params = {k: v for k, v in data.items() if v is not None}

        resp = await self.client.get(os.path.join(self.base, url), params=params)
        content = resp.read().strip(b"\n").strip(b" ")
        if content.startswith(b"{") and content.endswith(b"}"):
            return json.loads(content)
        elif content.startswith(b"[") and content.endswith(b"]"):
            return json.loads(content)
        else:
            return content.decode("utf-8")



import logging
from typing import Callable

from .message import ReflectionMessageBase
from ..enums import Platform, ChatType
from ..caching.base import MediaCache


class ReflectionAPI:
    def __init__(self, platform: Platform):
        self.logger = logging.getLogger(self.__class__.__name__)
        
        assert type(platform) == Platform
        self.platform = platform
        self.reflection_message_cls: type[ReflectionMessageBase] | None = None

        self.logger.debug(f"initializing ReflectionAPI for {self.platform}...")
        try:
            match platform:
                case Platform.DISCORD:
                    from .discord import DiscordReflectionAPI
                    self._ref = DiscordReflectionAPI()
                    from .message.discord import DiscordReflectionMessage
                    self.reflection_message_cls = DiscordReflectionMessage
                case Platform.TELEGRAM:
                    from .telegram import TelegramReflectionAPI
                    self._ref = TelegramReflectionAPI()
                    from .message.telegram import TelegramReflectionMessage
                    self.reflection_message_cls = TelegramReflectionMessage
                case _:
                    raise RuntimeError(f"ReflectionAPI unknown platform {self.platform}.")
        except ImportError:
            raise RuntimeError(
                f"ReflectionAPI for {self.platform} is not available, maybe you forgot to install its dependecies?")
        self.logger.debug(f"ReflectionAPI has been initilized for {self.platform}.")

    def run(self):
        return self._ref.run()

    def register_on_message_event(self, function: Callable):
        return self._ref.register_on_message_event(function)

    def register_slash_command(self, name: str, function: Callable, description: str | None = None):
        return self._ref.register_slash_command(name, function, description)

    def get_guild_info(self, guild) -> dict:
        return self._ref.get_guild_info(guild)

    def get_thread_info(self, thread) -> dict:
        return self._ref.get_thread_info(thread)

    def get_realtime_data(self, message: ReflectionMessageBase) -> dict:
        data = dict(
            platform=self.platform,
            chat_info=message.get_chat_info()
        )

        if guild_info := self.get_guild_info(message):
            data["guild_info"] = guild_info

        if thread_info := self.get_thread_info(message):
            data["thread_info"] = thread_info

        return data

    def can_read_history(self, channel) -> bool:
        return self._ref.can_read_history(channel)

    def __typecheck_reflectionmessage(self, message: ReflectionMessageBase):
        assert self.reflection_message_cls is not None
        assert isinstance(message, self.reflection_message_cls)

    def is_message_from_the_bot(self, message: ReflectionMessageBase) -> bool:
        self.__typecheck_reflectionmessage(message)
        return self._ref.is_message_from_the_bot(message)  # pyright: ignore[reportArgumentType]

    def is_bot_mentioned(self, message: ReflectionMessageBase) -> bool:
        self.__typecheck_reflectionmessage(message)
        return self._ref.is_bot_mentioned(message)  # pyright: ignore[reportArgumentType]

    async def is_dm_or_admin(self, message: ReflectionMessageBase) -> bool:
        self.__typecheck_reflectionmessage(message)
        return await self._ref.is_dm_or_admin(message)  # pyright: ignore[reportArgumentType]

    async def send_file(self, chat_id: int, filepath: str, filename: str, caption: str | None = None):
        return await self._ref.send_file(chat_id, filepath, filename, caption)  # pyright: ignore[reportArgumentType]

    async def get_user_avatar(self, user_id: int) -> MediaCache | None:
        return await self._ref.get_user_avatar(user_id)

    async def fetch_channel_history(self, channel_id: int, n: int = 10):
        return await self._ref.fetch_channel_history(channel_id, n)  # pyright: ignore[reportArgumentType]



from dataclasses import dataclass
import logging
import json
import time
from typing import Sequence

import httpx
from openai import AsyncOpenAI, APIError

from .enums import ChatRole
from .utils import CoroutineQueueExecutor, clean_dict, exists, format_time_ago
from .caching.base import MediaCache
from .typing import AsyncPredicate, AsyncFunction, Optional
from .reflection.message import ReflectionMessageBase
from .config import OpenAILanguageModelConfig, OpenAIAuthConfig


@dataclass(frozen=True)
class FunctionCall:
    name: str
    index: int
    id: str
    arguments: Optional[dict] = None

    def to_dict(self) -> dict:
        return dict(
            name=self.name,
            arguments=self.arguments,
            index=self.index,
            id=self.id
        )

    def to_openai_dict(self) -> dict:
        return dict(
            type="function",
            id=self.id,
            function=dict(
                name=self.name,
                arguments=json.dumps(self.arguments, ensure_ascii=False)
            )
        )

    @classmethod
    def from_dict(cls, data: dict) -> 'FunctionCall':
        return cls(
            name=data["name"],
            index=data["index"],
            id=data["id"],
            arguments=data.get("arguments", {}),
        )


@dataclass
class PartialFunctionCall:
    name: str
    index: int
    id: str
    arguments: Optional[str] = None

    def to_function_call(self) -> FunctionCall:
        return FunctionCall(
            name=self.name,
            index=self.index,
            id=self.id,
            arguments=json.loads(self.arguments) if self.arguments else {}
        )


class ChatMessage:
    def __init__(
        self,
        role: ChatRole | str,
        content: Optional[str] = None,
        metadata: Optional[dict] = None,
        message_time: Optional[float] = None,
        images: Optional[MediaCache | Sequence[MediaCache]] = None,
        audio: Optional[MediaCache | Sequence[MediaCache]] = None,
        tool_calls: Optional[list[FunctionCall]] = None,
        tool_call_id: Optional[str] = None,
        *,
        # TODO: add type hints and type checks
        instance_id: Optional[str] = None,
        origin: ReflectionMessageBase | None = None,
        bot=None
    ):
        self.instance_id = instance_id

        # the original message that this message is instantiated from, should be of type
        # discord.Message or telegram.Message, but the type is not specified here to avoid import
        # errors if one library is not found, is used mostly to reply to the original message
        self.origin = origin

        self.bot = bot

        assert role is not None, f"expected `role` to be of type `Role` and not be None, but got `{role}`"
        if images:
            from .caching import ImageCache

            assert isinstance(images, (ImageCache, list)
                              ), f"Images must be of type ImageCache or list[ImageCache], but got {images}."
            if isinstance(images, ImageCache):
                images = [images]
            else:
                for image in images:
                    assert isinstance(image, ImageCache), f"expected image to be ImageCache, but got {image}."
        else:
            images = []

        if audio:
            from .caching import AudioCache

            assert isinstance(audio, (AudioCache, list)
                              ), f"audio must be of type AudioCache or list[AudioCache], but got {audio}."
            if isinstance(audio, AudioCache):
                audio = [audio]
            else:
                for _audio in audio:
                    assert isinstance(_audio, AudioCache), f"expected audio to be of type AudioCache, but got {_audio}."
        else:
            audio = []

        # validating each role's requirements
        match role:
            case ChatRole.SYSTEM:
                assert exists(content, True) and isinstance(
                    content, str), f"expected SYSTEM to have a `content` of type `str` but got `{content}`"
                assert not exists(
                    metadata), f"expected SYSTEM to not have `metadata` (e.g. the metadata should be None) but got `{metadata}`"
                assert not exists(
                    tool_calls), f"expected SYSTEM to not have `tool_calls` (e.g. the tool_calls should be None) but got `{tool_calls}`"
                assert not exists(
                    tool_call_id), f"expected SYSTEM to not have `tool_call_id` (e.g. the tool_call_id should be None) but got `{tool_call_id}`"
                assert not exists(
                    images), f"Images can only be attached to user messages, but got {images} for role SYSTEM"

            case ChatRole.ASSISTANT:
                assert content is None or isinstance(
                    content, str), f"expected ASSISTANT to have a `content` of type `str` but got `{content}`"
                assert not exists(
                    metadata), f"expected ASSISTANT to not have `metadata` (e.g. the metadata should be None) but got `{metadata}`"
                assert not exists(
                    tool_call_id), f"expected ASSISTANT to not have `tool_call_id` (e.g. the tool_call_id should be None) but got `{tool_call_id}`"
                assert not exists(
                    images), f"Images can only be attached to user messages, but got {images} for role ASSISTANT"

                if exists(tool_calls):
                    assert not exists(
                        content, True), f"expected ASSISTANT to not have `content` (e.g. the content should be None) while `tool_calls` is not None but got `{content}`"
                    assert isinstance(
                        tool_calls, list), f"expected TOOL to have `tool_calls` of type `list[FunctionCall]` but got `{tool_calls}`"
                    for tc in tool_calls:
                        assert isinstance(
                            tc, FunctionCall), f"expected TOOL to have `tool_calls` of type `list[FunctionCall]` but at least one of the list elements is not of type FunctionCall, got `{tool_calls}`"

            case ChatRole.USER:
                assert exists(content, True) and isinstance(
                    content, str), f"expected USER to have a `content` of type `str` but got `{content}`"
                assert not exists(metadata) or isinstance(
                    metadata, dict), f"expected `metadata` to be None or `metadata` to be of type `dict` but got `{metadata}`"
                assert not exists(tool_calls) or (isinstance(tool_calls, list) and tool_calls == [
                ]), f"expected SYSTEM to not have `tool_calls` (e.g. the tool_calls should be None) but got `{tool_calls}`"
                assert not exists(
                    tool_call_id), f"expected USER to not have `tool_call_id` (e.g. the tool_call_id should be None) but got `{tool_call_id}`"

            case ChatRole.TOOL:
                assert exists(content, True) and isinstance(
                    content, str), f"expected TOOL to have a `content` of type `str` but got `{content}`"
                assert not exists(
                    metadata), f"expected TOOL to not have `metadata` (e.g. the metadata should be None) but got `{metadata}`"
                assert exists(tool_call_id) and isinstance(
                    tool_call_id, str), f"expected TOOL to have a `tool_call_id` of type `str` but got `{tool_call_id}`"
                assert not exists(
                    images), f"Images can only be attached to user messages, but got {images} for role TOOL"

            case _:
                raise ValueError(f"Invalid role \"{role}\".")

        self.role = role
        self.content = content
        self.metadata = metadata
        self.time = (message_time if message_time and message_time > 0 else time.time())

        self.images = images  # Should be an ImageCache instance or a list of ImageCache instances or None
        self.audio = audio

        # store function calls and function results
        self.tool_calls: list[FunctionCall] = tool_calls or []
        self.tool_call_id = tool_call_id

    @property
    async def instance(self):
        if not self.bot:
            return
        return await self.bot.get_conversation_instance(self.instance_id)

    def to_dict(self) -> dict:
        return clean_dict(dict(
            role=self.role,
            content=self.content,
            metadata=self.metadata,
            time=self.time,
            images=[x.to_dict() for x in self.images or []],
            audio=[x.to_dict() for x in self.audio or []],
            tool_calls=[x.to_dict() for x in self.tool_calls or []],
            tool_call_id=self.tool_call_id
        ))

    @classmethod
    def from_dict(cls, data: dict) -> 'ChatMessage':
        images = []
        if images_cache := data.get("images", []):
            from .caching import ImageCache
            images = [ImageCache.from_dict(d) for d in images_cache]
        audio = []
        if audio_cache := data.get("audio", []):
            from .caching import AudioCache
            audio = [AudioCache.from_dict(d) for d in audio_cache]
        return cls(
            role=ChatRole[data["role"].upper()],
            content=data.get("content"),
            metadata=data.get("metadata"),
            message_time=data.get("time", time.time()),
            images=images,
            audio=audio,
            tool_calls=[FunctionCall.from_dict(d) for d in data.get("tool_calls", [])],
            tool_call_id=data.get("tool_call_id")
        )

    def to_openai_dict(self, timestamps: bool = True) -> dict:
        openai_dict = dict(role=self.role.lower())  # ensure role is in lower case
        match self.role:
            case ChatRole.USER:
                content = [f"User: {self.content}"]
                if timestamps:
                    fmt_time = time.strftime("%Y-%m-%d %H:%M:%S %Z", time.localtime(self.time))
                    fmt_diff_time = format_time_ago(time.time()-self.time)
                    content.append(f"Sent At: {fmt_time} ({fmt_diff_time})")
                if self.metadata is not None:
                    content += [
                        "Metadata:",
                        ""
                        "```json",
                        json.dumps(self.metadata, ensure_ascii=False),
                        "```"
                    ]
                content_dict: list[dict] = [dict(type="text", text="\n".join(content))]
                for img in self.images:
                    content_dict.append(dict(
                        type="image_url",
                        image_url=dict(url=img.to_data_url())
                    ))
                for audio in self.audio:
                    content_dict.append(dict(
                        input_audio=dict(
                            data=audio.to_base64(),
                            format="wav"  # for some reason this value is not used by the API is required to be mp3 or wav
                        ),
                        type="input_audio",
                    ))
                openai_dict.update(dict(content=content_dict))  # type: ignore
            case ChatRole.ASSISTANT:
                if exists(self.content, True):
                    openai_dict.update(dict(content=self.content))  # type: ignore
                else:
                    openai_dict.update(dict(
                        tool_calls=[i.to_openai_dict() for i in self.tool_calls]
                    ))  # type: ignore
            case ChatRole.TOOL:
                openai_dict.update(dict(
                    content=self.content,
                    tool_call_id=self.tool_call_id
                ))  # type: ignore
            case _:
                openai_dict.update(dict(content=self.content))  # type: ignore
        return clean_dict(openai_dict)


# takes the last N messages in the message list that match a predicate, and moves them to the
# end of the message list, this is to ensure the LLMs immmediate context is accurate
async def get_rearranged_messages(messages: list[ChatMessage], predicate: AsyncPredicate, n: int = 5) -> list[ChatMessage]:
    selected_messages: list[ChatMessage] = []
    unselected_predicate: list[ChatMessage] = []
    for msg in messages[::-1]:
        if await predicate(msg) and len(selected_messages) < n:
            selected_messages.append(msg)
        else:
            unselected_predicate.append(msg)

    return unselected_predicate[::-1] + selected_messages[::-1]


class AsyncChatClient:
    def __init__(
        self,
        auth: OpenAIAuthConfig,
        model: OpenAILanguageModelConfig,
        messages: Optional[list[ChatMessage]] = None,
        log_tool_calls: bool = False,
        enforce_system_header: bool = False,
    ):
        self.logger = logging.getLogger(self.__class__.__name__)

        self.auth = auth
        self.model = model
        self.messages = messages or []
        self.log_tool_calls = log_tool_calls
        self.enforce_system_header = enforce_system_header

        if messages is not None:
            assert isinstance(messages, list), \
                f"expected messages to be of type `list[ChatMessage]` or be None, but got `{messages}`"
            for m in messages:
                assert isinstance(m, ChatMessage), \
                    f"expected messages to be of type `list[ChatMessage]` or be None, but at least one element in the list is not of type `ChatMessage`, got `{messages}`"

        self.system_prompt = None
        self.tools: dict = {}
        self.rearrange_predicate = None

    # Hotfix: create a new session every time to avoid issues with cancelling requests and connection errors

    @property
    def session(self):
        return AsyncOpenAI(
            api_key=self.auth.api_key,
            base_url=self.auth.base_url,
            # don't use proxies from environment variables
            http_client=httpx.AsyncClient(trust_env=False),
        )

    def add_tool(self, name: str, func: AsyncFunction, parameters: Optional[dict] = None, description: Optional[str] = None):
        """
        Register a tool (function) for tool calling.
        name: tool name (string)
        func: (str) -> Awaitable[None]
        parameters: OpenAI tool/function parameters schema (dict)
        description: description of the tool (string)
        """
        self.tools[name] = dict(
            func=func,
            schema=dict(
                type="function",
                function=dict(
                    name=name,
                    description=description or name,
                    parameters=parameters or dict(),
                )
            )
        )

    def set_rearrange_predicate(self, predicate: AsyncPredicate):
        """
        Set a predicate function to rearrange messages based on a specific condition.
        This is useful for filtering messages by channel or other criteria and moving them
        to the front of the model context.
        """

        self.rearrange_predicate = predicate

    def set_system(self, prompt: str):
        assert prompt is not None and prompt != ""
        self.system_prompt = prompt

    def add_message(self, message: ChatMessage | str) -> ChatMessage:
        assert message is not None
        if isinstance(message, str):
            role = None
            match self.messages[-1].role:
                case ChatRole.SYSTEM:
                    role = ChatRole.ASSISTANT
                case ChatRole.ASSISTANT:
                    role = ChatRole.USER
                case ChatRole.USER:
                    role = ChatRole.ASSISTANT
                case _:
                    role = ChatRole.ASSISTANT
            message = ChatMessage(role, message)
        elif not isinstance(message, ChatMessage):
            raise ValueError(f"expected message to be a ChatMessage or a string, but got {message}.")
        self.messages.append(message)
        return message

    def set_messages(self, messages: list[ChatMessage]):
        assert isinstance(messages, list), f"Invalid messages \"{messages}\"."
        self.messages = messages

    def get_messages(self):
        return self.messages

    async def create_chat_completion(self, stream: bool = False, enable_timestamps: bool = True):
        openai_messages: list = await self.get_openai_messages_dict(enable_timestamps=enable_timestamps)
        kwargs = dict(
            model=self.model.id,
            messages=openai_messages,
            temperature=0.7,
            top_p=0.9,
            max_tokens=1024,
            parallel_tool_calls=True,
            # reasoning_effort="high",
            stream=stream,
        )
        # If tools are registered, add them to the request
        if self.tools:
            kwargs["tools"] = [v["schema"] for k, v in self.tools.items()]  # type: ignore

        # Debugging: print the request to a file
        # with open("last-request.json", "w", encoding="utf-8") as f:
        #     json.dump(kwargs, f, ensure_ascii=False, indent=4)

        return await self.session.chat.completions.create(**kwargs)  # type: ignore

    async def get_openai_messages_dict(self, enable_timestamps: bool = True) -> list[dict]:
        if len(self.messages) == 0:
            return list()

        messages = self.enforce_approximate_context_limit(self.messages.copy())

        openai_messages = [msg.to_openai_dict(timestamps=enable_timestamps) for msg in messages]
        # add system prompt before to the last user message message to ensure it is in the
        # model's context but also don't disrupt the conversation or tool calling
        if exists(self.system_prompt):
            system_message = ChatMessage(
                role=ChatRole.SYSTEM,
                content=self.system_prompt
            ).to_openai_dict(timestamps=enable_timestamps)

            insert_index = 0
            # some models only support system message as the very first message
            if not self.enforce_system_header:
                for i, message in enumerate(messages):
                    if message.role == ChatRole.USER:
                        insert_index = i
            openai_messages.insert(insert_index, system_message)
        return openai_messages

    def enforce_approximate_context_limit(self, messages: list[ChatMessage]):
        system_prompt_size = len(self.system_prompt) if self.system_prompt else 0
        request_size = system_prompt_size
        for cutoff_index, message in enumerate(messages):
            role = message.role
            context = message.content or ""
            # TODO: better handle images and audio
            request_size += len(role) + len(context) + 1024 * len(message.audio) + 1024 * len(message.images)
            # approximate size of the message in tokens, assuming 4 characters per token
            if request_size > self.model.max_context * 4:
                break
        else:
            cutoff_index = None

        if cutoff_index:
            if cutoff_index == 0:
                self.logger.warning("No messages fit in the request, cutting off the first message.")
                message_cut = messages[0]
                assert message_cut.content, "Message content must not be empty, this is a bug."
                message_cut.content = "(This message was cut off due to length limitations)\n\n" + \
                    message_cut.content[:request_size-system_prompt_size -
                                        200]  # cut off the message to fit in the request
            self.logger.warning("unable to fit all messages in one request.")
            messages = messages[-(cutoff_index-1):]
        return messages

    async def stream_completion(self, start_think_tag: str = "<think>", end_think_tag: str = "</think>"):
        lookbehind_buffer: str = ""
        max_buffer_size = max(len(start_think_tag), len(end_think_tag))

        is_thinking = False

        async for char in self.stream_request():
            lookbehind_buffer += char

            # make sure buffer doesn't grow indefinitely if none of the tags are found
            if len(lookbehind_buffer) > max_buffer_size:
                char_to_yield = lookbehind_buffer[0]
                lookbehind_buffer = lookbehind_buffer[1:]
                if not is_thinking:
                    yield char_to_yield

            # check for tags
            if lookbehind_buffer.endswith(start_think_tag) and not is_thinking:
                is_thinking = True
                lookbehind_buffer = ""
            elif lookbehind_buffer.endswith(end_think_tag) and is_thinking:
                is_thinking = False
                lookbehind_buffer = ""

        if not is_thinking and lookbehind_buffer:
            for c in lookbehind_buffer:
                yield c

    async def get_tool_results(self, function_calls: list[FunctionCall], reference_message: ChatMessage):
        """
        Execute tool calls and return results
        """

        results = [ChatMessage(
            ChatRole.ASSISTANT,
            content=None,
            tool_calls=function_calls
        )]
        for fn in function_calls:
            tool = self.tools.get(fn.name)
            if tool:
                func = tool["func"]
                func_args_str = ""
                if fn.arguments:
                    func_args_str = ", ".join([
                        (f"{k}=\"{v}\"" if isinstance(v, str) else f"{k}={v}")
                        for k, v in fn.arguments.items()
                    ])
                try:
                    if self.log_tool_calls:
                        self.logger.info(f"Calling {fn.name}({func_args_str})...")
                    else:
                        self.logger.debug(f"Calling {fn.name}({func_args_str})...")
                    result = await func(reference_message, **fn.arguments)  # type: ignore
                    result = json.dumps(result, ensure_ascii=False, indent=2)
                except Exception as e:
                    self.logger.exception(f"Error calling tool '{fn.name}({func_args_str})'")
                    result = f"Tool error: {e}"

                if self.log_tool_calls:
                    self.logger.info(f"Call {fn.name}({func_args_str}): {result}")
                else:
                    self.logger.debug(f"Call {fn.name}({func_args_str}): {result}")
            else:
                result = f"Tool '{fn.name}' was not found."
                self.logger.warning(f"Tool '{fn.name}' was not found.")

            results.append(ChatMessage(
                ChatRole.TOOL,
                content=result,
                tool_call_id=fn.id,
            ))
        return results

    async def stream_request(
        self,
        verbose: bool = logging.getLevelName(logging.root.level).lower() == "debug",
        reference_message: ChatMessage | None = None
    ):
        # reference_message is the last user message in which we are responding to, it contains
        # information about the current channel and holds special metadata
        reference_message = reference_message or self.messages[-1]
        assert reference_message.role == ChatRole.USER, f"`reference_message` must be a USER message, but got a message with the role of \"{reference_message.role}\"."

        async def execute_tool_task(function_call):
            self.messages += await self.get_tool_results(
                [function_call],
                reference_message=reference_message
            )

        async with CoroutineQueueExecutor() as tool_call_queue:
            partial_function_calls: dict[int, PartialFunctionCall] = dict()
            last_function_call_index = None
            response = ""
            finish_reason = None
            async for event in await self.create_chat_completion(stream=True):
                if event.choices is None or len(event.choices) == 0:
                    continue
                choice = event.choices[0]

                if exists(_tool_calls := choice.delta.tool_calls):
                    for tool_call in _tool_calls:
                        function = tool_call.function
                        function_index = tool_call.index
                        function_id = tool_call.id
                        function_name = function.name
                        function_arguments = function.arguments
                        if prev_function_call := partial_function_calls.get(function_index):
                            if function_arguments:
                                new_function_arguments = (prev_function_call.arguments or "") + function_arguments
                            else:
                                new_function_arguments = prev_function_call.arguments
                            partial_function_calls[function_index] = PartialFunctionCall(
                                name=prev_function_call.name,
                                arguments=new_function_arguments,
                                index=function_index,
                                id=prev_function_call.id
                            )
                        else:
                            if verbose:
                                print(f"Function: {function_name}", end="")
                            partial_function_calls[function_index] = PartialFunctionCall(
                                name=function_name,
                                arguments=function_arguments,
                                index=function_index,
                                id=function_id
                            )
                        # if the model is producing a new function call, we can assume the function call is finished generating
                        if last_function_call_index is not None and function_index != last_function_call_index:
                            function_call = partial_function_calls.pop(function_index).to_function_call()
                            # runs the tool call without blocking and in correct order
                            await tool_call_queue.add_to_queue(execute_tool_task(function_call))
                        last_function_call_index = function_index
                        if verbose:
                            print(function_arguments, end="")

                if choice.finish_reason:
                    finish_reason = choice.finish_reason

                if content := choice.delta.content:
                    if verbose:
                        print(content, end="", flush=True)
                    response += content
                    # now yields the response character by character for easier parsing
                    for char in content:
                        yield char

                if verbose and hasattr(choice.delta, "reasoning_content") and (reasoning_content := choice.delta.reasoning_content):
                    print(reasoning_content, end="", flush=True)

            if verbose:
                print()

        function_calls = [f.to_function_call() for f in partial_function_calls.values()]
        del partial_function_calls
        if function_calls:
            self.messages += await self.get_tool_results(
                function_calls,
                reference_message=reference_message
            )

        if response:
            self.messages.append(ChatMessage(
                role=ChatRole.ASSISTANT,
                content=response
            ))

        if finish_reason == "tool_calls":
            # Recursively call stream_request and yield its results
            async for chunk in self.stream_request(reference_message=reference_message):
                yield chunk

    async def stream_ask(self, message: str | ChatMessage, temporal: bool = False):
        if temporal:
            orig_messages = self.messages.copy()

        if isinstance(message, str):
            message = ChatMessage(ChatRole.USER, message)
        else:
            assert isinstance(message, ChatMessage), "Message must be a string or a ChatMessage."

        assert message.role == ChatRole.USER, "Message must be from the user."
        self.messages.append(message)

        async for chunk in self.stream_completion():
            yield chunk

        if temporal:
            self.messages = orig_messages  # type: ignore

    def state_dict(self):
        return dict(
            messages=[msg.to_dict() for msg in self.messages]
        )

    def load_state_dict(self, data: dict):
        self.messages = [ChatMessage.from_dict(msg) for msg in data.get("messages") or []]



import asyncio
from enum import StrEnum
import logging
import os
from pathlib import Path
from types import CoroutineType
from typing import IO, Sequence


class CoroutineQueueExecutor:
    def __init__(self, max_queue_size: int = 1000):
        self.logger = logging.getLogger(self.__class__.__name__)

        self.max_queue_size = max_queue_size
        self.tasks: list[CoroutineType] = []
        self.execute_lock = asyncio.Lock()
        self.execute_task = None

    async def __execute_queue(self):
        await self.execute_lock.acquire()
        while len(self.tasks) > 0:
            await self.tasks[0]
            del self.tasks[0]
        self.execute_lock.release()

    async def add_to_queue(self, t: CoroutineType):
        self.tasks.append(t)
        # if execute_lock is not acquired, we are not executing anything
        # in that case we should start executing the tasks
        if not self.execute_lock.locked():
            self.execute_task = asyncio.create_task(self.__execute_queue())

    async def __aenter__(self):
        self.execute_task = None
        if self.execute_lock.locked():
            self.execute_lock.release()

        if len(self.tasks) > 0:
            self.logger.warning(f"Corotine never avaited: {self.tasks}")

        self.tasks = []

        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self.execute_task:
            await self.execute_task

# helpers


def exists(value, allow_empty_string=False):
    if value is None:
        return False

    if isinstance(value, (int, float)):
        return True

    if isinstance(value, str):
        return value != "" or (value == "" and allow_empty_string)

    if isinstance(value, (list, tuple, dict)):
        return len(value) != 0

    return True


def load_dotenv():
    # load environment variables
    try:
        import dotenv  # pyright: ignore[reportMissingImports]
    except ImportError:
        logging.warning("dotenv is not installed, install it using `pip install dotenv`")
    else:
        dotenv.load_dotenv()


# modified from https://stackoverflow.com/a/76636817
def format_time_ago(delta: float, min_count: int = 3, max_lenght: int = 2) -> str:
    """Return time difference as human-readable string"""

    periods = (
        ("year", 31536000),
        ("month", 2592000),
        ("week", 604800),
        ("day", 86400),
        ("hour", 3600),
        ("minute", 60),
        ("second", 1),
    )

    fmt_list = []
    for period, seconds_each in periods:
        if delta >= seconds_each:
            how_many = int(delta / seconds_each)
            if how_many >= min_count:
                fmt_list.append(f"{how_many} {period}{'s' if how_many >= 2 else ''}")
                if len(fmt_list) >= max_lenght:
                    break

                delta -= seconds_each * how_many

    if fmt_list:
        return " and ".join(fmt_list) + " ago"

    return "just now"  # less than a second ago


class Ansi(StrEnum):
    END = '\33[0m'
    BOLD = '\33[1m'
    ITALIC = '\33[3m'
    URL = '\33[4m'
    BLINK = '\33[5m'
    BLINK2 = '\33[6m'
    SELECTED = '\33[7m'

    BLACK = '\33[30m'
    RED = '\33[31m'
    GREEN = '\33[32m'
    YELLOW = '\33[33m'
    BLUE = '\33[34m'
    VIOLET = '\33[35m'
    BEIGE = '\33[36m'
    WHITE = '\33[37m'

    BLACKBG = '\33[40m'
    REDBG = '\33[41m'
    GREENBG = '\33[42m'
    YELLOWBG = '\33[43m'
    BLUEBG = '\33[44m'
    VIOLETBG = '\33[45m'
    BEIGEBG = '\33[46m'
    WHITEBG = '\33[47m'

    GREY = '\33[90m'
    RED2 = '\33[91m'
    GREEN2 = '\33[92m'
    YELLOW2 = '\33[93m'
    BLUE2 = '\33[94m'
    VIOLET2 = '\33[95m'
    BEIGE2 = '\33[96m'
    WHITE2 = '\33[97m'

    GREYBG = '\33[100m'
    REDBG2 = '\33[101m'
    GREENBG2 = '\33[102m'
    YELLOWBG2 = '\33[103m'
    BLUEBG2 = '\33[104m'
    VIOLETBG2 = '\33[105m'
    BEIGEBG2 = '\33[106m'
    WHITEBG2 = '\33[107m'


def get_resource_path(filename) -> str:
    return os.path.join(PixiPaths.resources(), filename)


def open_resource(filename: str, mode: str) -> IO:
    return open(os.path.join(PixiPaths.resources(), filename), mode=mode, encoding="utf-8")


class PixiPaths:
    _root = Path("~/.pixi")

    # ---- configuration ----
    @classmethod
    def set_root(cls, root: str | Path) -> None:
        cls._root = Path(root)

    @classmethod
    def root(cls) -> Path:
        return cls._root.expanduser()

    # ---- paths ----
    @classmethod
    def addons(cls) -> Path:
        return cls.root() / "addons"

    @classmethod
    def datasets(cls) -> Path:
        return cls.root() / "datasets"

    @classmethod
    def resources(cls) -> Path:
        return cls.root() / "resources"

    @classmethod
    def userdata(cls) -> Path:
        return cls.root() / "userdata"

    @classmethod
    def cache(cls) -> Path:
        return cls.root() / "cache"


# if the PixiPaths.RESOURCES folder doesn't exist, we should copy all our default assets in there when the module is imported
if __package__ is not None:
    import importlib.resources
    import shutil

    MODULE_PATH = importlib.resources.files(__package__)
    RESOURCES_PATH = str(MODULE_PATH / "resources")

    def copy_if_absent(src: str, dst: str, *, follow_symlinks: bool = True):
        if os.path.exists(dst):
            if os.path.isdir(dst):
                raise FileExistsError(f"directory exists with the same name as destination the file: {dst}")
            return
        shutil.copy2(src, dst, follow_symlinks=follow_symlinks)

    def copy_default_resources():
        os.makedirs(PixiPaths.resources(), exist_ok=True)
        shutil.copytree(RESOURCES_PATH, PixiPaths.resources(), dirs_exist_ok=True, copy_function=copy_if_absent)


def clean_dict(d: dict):
    final_dict = {}
    for k, v in d.items():
        if v is None:
            continue
        elif isinstance(v, dict):
            if len(v) == 0:
                continue
            v = clean_dict(v)
        elif isinstance(v, Sequence):
            if len(v) == 0:
                continue
            if not isinstance(v, str): # other than strings, we every other sequence is a nested dynamic type
                v = [(clean_dict(e) if isinstance(e, dict) else e) for e in v]
        final_dict[k] = v
    return final_dict



import dataclasses
from typing import Any

from ...caching.base import MediaCache
from ...enums import Platform, ChatType, Messages


@dataclasses.dataclass(frozen=True)
class ReflectionMessageAuthor:
    id: int
    first_name: str
    last_name: str
    display_name: str
    mention: str


@dataclasses.dataclass(frozen=True)
class ReflectionEnvironment:
    chat_id: int
    chat_title: str
    chat_type: ChatType
    forum_id: int
    is_forum: bool


@dataclasses.dataclass(frozen=True)
class ReflectionMessageBase:
    content: str
    author: ReflectionMessageAuthor
    id: int
    environment: ReflectionEnvironment
    platform: Platform

    # hold a refrence of the original message for everything else that we didn't define here
    origin: Any

    @classmethod
    def from_origin(cls, message) -> 'ReflectionMessageBase':
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("__to_reflectionmessage", cls.__name__))

    async def send(self, content: str) -> 'ReflectionMessageBase':
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("send", self.__class__.__name__))

    async def edit(self, content: str) -> 'ReflectionMessageBase | None':
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("edit", self.__class__.__name__))

    async def delete(self):
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("delete", self.__class__.__name__))

    async def typing(self):
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("typing", self.__class__.__name__))

    async def fetch_images(self) -> list[MediaCache]:
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("fetch_images", self.__class__.__name__))

    async def fetch_audio(self) -> list[MediaCache]:
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("fetch_audio", self.__class__.__name__))

    async def fetch_refrences(self) -> 'ReflectionMessageBase | None':
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("fetch_message_reply", self.__class__.__name__))

    async def add_reaction(self, emoji: str):
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("add_reaction", self.__class__.__name__))

    async def send_file(self, filepath: str, filename: str, caption: str | None = None):
        raise NotImplementedError(Messages.NOT_IMPLEMENTED % ("send_file", self.__class__.__name__))
    
    @property
    def environment_id(self) -> str:
        # if the message is in a forum type environemnt (server, guild, channels group, etc.)
        if self.environment.is_forum:
            return f"forum#{self.environment.forum_id}"
        return f"chat#{self.environment.chat_id}"

    def get_chat_info(self) -> dict:
        return {"type": self.environment.chat_type, "name": self.environment.chat_title, "id": self.environment.chat_id}
    
    def is_inside_dm(self) -> bool:
        return self.environment.chat_type == ChatType.PRIVATE


import logging
import discord

from . import ReflectionMessageBase, ReflectionMessageAuthor, ReflectionEnvironment, ChatType, Platform
from ...caching.base import MediaCache

ImageCache = None
AudioCache = None
try:
    from ...caching import ImageCache
except ImportError:
    pass
try:
    from ...caching import AudioCache
except ImportError:
    pass


class DiscordReflectionMessage(ReflectionMessageBase):
    origin: discord.Message | discord.Interaction

    @classmethod
    def from_origin(cls, message: discord.Message | discord.Interaction) -> 'DiscordReflectionMessage':
        user = None
        content = ""
        if isinstance(message, discord.Message):
            user = message.author
            content = message.content if message.content else ""
        elif isinstance(message, discord.Interaction):
            user = message.user
        assert user is not None
        channel = message.channel
        chat_title = ""
        if channel and hasattr(channel, "name"):
            chat_title = (channel.name or "")  # pyright: ignore[reportAttributeAccessIssue]
        chat_type = ChatType.UNKNOWN
        match channel.type if channel else None:
            case discord.ChannelType.private:
                chat_type = ChatType.PRIVATE
            case discord.ChannelType.group:
                chat_type = ChatType.GROUP
            case discord.ChannelType.text:
                chat_type = ChatType.TEXT

        environment = ReflectionEnvironment(
            chat_id=channel.id if channel else -1,
            chat_title=chat_title,
            chat_type=chat_type,
            forum_id=message.guild.id if message.guild else -1,
            is_forum=message.guild is not None
        )
        return DiscordReflectionMessage(
            content=content,
            author=ReflectionMessageAuthor(
                id=user.id,
                first_name=user.name,
                last_name="",
                display_name=user.display_name,
                mention=user.mention,
            ),
            id=message.id,
            environment=environment,
            platform=Platform.DISCORD,
            origin=message
        )

    async def send(self, content: str) -> 'DiscordReflectionMessage':
        if isinstance(self.origin, discord.Message):
            return self.from_origin(await self.origin.channel.send(content))
        elif isinstance(self.origin, discord.Interaction):
            await self.origin.response.send_message(content)
            return self.from_origin(await self.origin.original_response())
        else:
            raise TypeError(f"unknown origin of type {type(self.origin)}")

    async def edit(self, content: str) -> 'DiscordReflectionMessage':
        if isinstance(self.origin, discord.Message):
            return self.from_origin(await self.origin.edit(content=content))
        elif isinstance(self.origin, discord.Interaction):
            await self.origin.response.edit_message(content=content)
            return self.from_origin(await self.origin.original_response())
        else:
            raise TypeError(f"unknown origin of type {type(self.origin)}")

    async def delete(self):
        if isinstance(self.origin, discord.Message):
            await self.origin.delete()
        elif isinstance(self.origin, discord.Interaction):
            origin = await self.origin.original_response()
            await origin.delete()
        else:
            raise TypeError(f"unknown origin of type {type(self.origin)}")

    async def typing(self):
        channel = self.origin.channel
        if channel and hasattr(channel, "typing"):
            return await channel.typing()  # pyright: ignore[reportAttributeAccessIssue]

    async def fetch_images(self) -> list[MediaCache]:
        if isinstance(self.origin, discord.Interaction):
            return []

        supported_mime_types = {'image/jpeg', 'image/png', 'image/webp'}
        attachments = []
        for attachment in self.origin.attachments:
            if attachment.content_type in supported_mime_types:
                assert ImageCache
                attachments.append(ImageCache(await attachment.read()))
        return attachments

    async def fetch_audio(self) -> list[MediaCache]:
        if isinstance(self.origin, discord.Interaction):
            return []

        supported_mime_types = {'audio/mp3', 'audio/aac', 'audio/ogg', 'audio/flac', 'audio/opus'}
        supported_extensions = {'.mp3', '.aac', '.ogg', ".flac", ".opus", ".wav", ".webm", ".m4a"}
        attachments = []
        for attachment in self.origin.attachments:
            mime = attachment.content_type
            ext = attachment.filename.lower().rsplit('.', 1)[-1] if '.' in attachment.filename else ''
            ext = f'.{ext}'
            if (mime and mime in supported_mime_types) or (ext in supported_extensions):
                assert AudioCache
                attachments.append(AudioCache(await attachment.read()))
        return attachments

    async def fetch_refrences(self) -> 'DiscordReflectionMessage | None':
        origin: discord.Message | discord.Interaction = self.origin
        if isinstance(origin, discord.Interaction):
            return
        ref = origin.reference
        if ref is None:
            return
        if ref.cached_message:
            return DiscordReflectionMessage.from_origin(ref.cached_message)
        if ref.message_id is None:
            return
        return DiscordReflectionMessage.from_origin(await origin.channel.fetch_message(ref.message_id))

    async def add_reaction(self, emoji: str):
        origin: discord.Message | discord.Interaction = self.origin
        if isinstance(origin, discord.Interaction):
            return
        try:
            await origin.add_reaction(emoji)
        except discord.errors.NotFound:
            logging.exception("unable to add reaction, message not found.")
        except discord.Forbidden:
            logging.exception("unable to add reaction, operation forbidden.")
        except Exception:
            logging.exception("unknown error while adding a reaction to a message")

    async def send_file(self, filepath: str, filename: str, caption: str | None = None):
        try:
            channel = self.origin.channel
            if channel and hasattr(channel, "send"):
                with open(filepath, "rb") as f:
                    await channel.send(  # pyright: ignore[reportAttributeAccessIssue]
                        content=caption,
                        file=discord.File(fp=f, filename=filename)
                    )
        except discord.Forbidden:
            logging.exception("unable to send file, operation forbidden.")
        except discord.HTTPException as e:
            logging.exception(f"HTTPException while sending file: {e}")
        except Exception:
            logging.exception("unknown error while sending a file")



from dataclasses import dataclass
import logging
from typing import Optional

from .api import APIBase


@dataclass
class WikiMediaSearchResult:
    title: str
    snippet: Optional[str] = None
    url: Optional[str] = None
    description: Optional[str] = None
    pageid: Optional[str] = None


class AsyncWikimediaAPI(APIBase):
    def __init__(self, base_url: str):
        super().__init__(base=base_url, api_key=None, api_key_env_var=None)

    async def apiphp_request(self, **kwargs) -> dict | list:
        data = kwargs | dict(format="json", formatversion=2, redirects="resolve")
        return await self.request(
            "api.php",
            data=data
        )  # type: ignore

    async def search(self, srsearch: str) -> list[WikiMediaSearchResult]:
        resp = await self.apiphp_request(
            action="query",
            origin="*",
            generator="search",
            gsrsearch=srsearch,
            gsrlimit=5,
            gsrinfo="",
            gsrprop="snippet",
            prop="extracts|info",
            inprop="url",
            exintro=1,
            explaintext=1,
            exchars=500,
            exlimit=3
        )
        
        # HOTFIX: sometimes the api just fails, this could be cloudflare blocking our
        # IP address (e.g. minecraft wiki as of july 11th, 2025)
        if isinstance(resp, str):
            logging.warning("API FAILURE: the responce from MediaWiki API contains an invalid json object.")
            return [WikiMediaSearchResult(
                title="API FAILURE",
                snippet=None,
                description="the responce from MediaWiki API contains an invalid json object.",
                pageid=None,
                url=None
            )]

        results = []
        for page_data in resp.get("query", {}).get("pages", []):  # type: ignore
            title = page_data.get("title")
            snippet = page_data.get("snippet")
            extract = page_data.get("extract")
            pageid = page_data.get("pageid")
            fullurl = page_data.get("fullurl")
            assert title
            results.append(WikiMediaSearchResult(
                title=title,
                snippet=snippet,
                description=extract,
                pageid=pageid,
                url=fullurl
            ))

        return results

    async def opensearch(self, search: str) -> list[WikiMediaSearchResult]:
        result = await self.apiphp_request(
            action="opensearch",
            search=search,
        )
        print(result)
        results = []
        for title, desc, url in zip(result[1], result[2], result[3]):
            results.append(WikiMediaSearchResult(title=title, url=url, description=desc or None))

        return results

    async def get_info(self, titles: Optional[str] = None, pageids: Optional[str] = None):
        return await self.apiphp_request(
            action="query",
            titles=titles,
            pageids=pageids,
            prop="extracts|description|info",
            list="",
            redirects=1,
            exsentences=10,
            exintro=1,
            explaintext=1,
            inprop="url"
        )

    async def get_page(self, page: Optional[str] = None, pageid: Optional[str] = None):
        return await self.apiphp_request(
            action="parse",
            page=page,
            pageid=pageid,
            prop="wikitext",
            contentmodel="wikitext"
        )
    
    async def get_plaintext(self, title: str) -> tuple[str, str]:
        result: dict = await self.apiphp_request(
            action="query",
            prop="extracts",
            explaintext=1,
            format="json",
            titles=title
        ) # pyright: ignore[reportAssignmentType]
        pagecontent = result["query"]["pages"][0]
        return pagecontent["extract"], pagecontent["title"]



from dataclasses import asdict
import asyncio
import json
import logging
import math
import time

from .typing import AsyncPredicate, Optional
from .chatbot import AsyncChatbotInstance, CachedAsyncChatbotFactory, PredicateCommand, PredicateTool
from .agents import AgentBase, RetrievalAgent
from .apis import AsyncTenorAPI, AsyncWikimediaAPI
from .caching import SUPPORTS_MEDIA_CACHING
from .chatting import ChatMessage
from .enums import ChatRole, Messages, Platform
from .reflection import ReflectionAPI, ReflectionMessageBase
from .addon import AddonManager
from .config import OpenAIAuthConfig, OpenAIEmbeddingModelConfig, OpenAILanguageModelConfig, PixiFeatures, IdFilter
from .database import AsyncEmbeddingDatabase


# constants

COMMAND_PREFIXES = ["!pixi", "!pix", "!p"]
COMMAND_KEYWORDS = ["pixi", "پیکسی"]


# helper functions

def remove_prefixes(text: str):
    for prefix in COMMAND_PREFIXES:
        text = text.removeprefix(prefix)
    return text


class PixiClient:
    def __init__(
        self,
        platform: Platform,
        *,
        auth: OpenAIAuthConfig,
        model: OpenAILanguageModelConfig,
        helper_model: OpenAILanguageModelConfig | None,
        embedding_model: OpenAIEmbeddingModelConfig | None,
        features: PixiFeatures = PixiFeatures.EnableToolCalling | PixiFeatures.EnableToolLogging,
        database_names: Optional[list[str]] = None,
        environment_filter: IdFilter = IdFilter.allow(),
    ):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info(f"Initalizing Pixi with feature set {features} and platform {platform}...")

        self.platform = platform

        self.auth = auth
        self.model = model
        self.helper_model = helper_model or model
        self.embedding_model = embedding_model
        self.database_names = database_names or []
        self.database_tools_initalized = asyncio.Event()
        self.features = features
        self.enable_tool_calls = PixiFeatures.EnableToolCalling in features
        self.log_tool_calls = PixiFeatures.EnableToolLogging in features
        self.environment_filter = environment_filter

        self.accept_images = PixiFeatures.EnableImageSupport in features
        self.accept_audio = PixiFeatures.EnableAudioSupport in features
        if (self.accept_images or self.accept_audio) and not SUPPORTS_MEDIA_CACHING:
            self.logger.warning("tried to accept audio/images, but media caching features are not available")
            self.accept_images = False
            self.accept_audio = False

        self.chatbot_factory = CachedAsyncChatbotFactory(
            parent=self,
            auth=self.auth,
            model=model,
            hash_prefix=platform,
            log_tool_calls=self.log_tool_calls,
        )

        self.reflection_api = ReflectionAPI(platform=platform)

        self.gif_api = None
        if PixiFeatures.EnableGIFSearch in features:
            try:
                self.gif_api = AsyncTenorAPI()
            except KeyError:
                self.logger.warning("TENOR_API_KEY is not set, TENOR API features will not be available.")

        self.__register_builtin_commands()
        
        # TODO: add configurable wikis
        if self.enable_tool_calls and PixiFeatures.EnableWikiSearch in features:
            self.register_mediawiki_tools(url="https://minecraft.wiki/", wiki_name="minecraft")
            self.register_mediawiki_tools(url="https://www.wikipedia.org/w/", wiki_name="wikipedia")
            # self.register_mediawiki_tools(url="https://mcdf.wiki.gg/", wiki_name="minecraft_discontinued_features")

        if platform == Platform.DISCORD and self.enable_tool_calls:
            self.__register_discord_specific_tools()

        self.__register_builtin_slash_commands()

        self.addon_manager = AddonManager(self)
        self.addon_manager.load_addons()

        self.reflection_api.register_on_message_event(self.on_message)

        self.logger.info(f"Pixi finised initalizing!")

    def register_tool(self, name: str, func, parameters: dict, description: Optional[str], predicate: Optional[AsyncPredicate] = None):
        if not self.enable_tool_calls:
            self.logger.warning("tried to register a tool, but tool calls are disabled, ignoring...")
            return

        self.chatbot_factory.register_tool(PredicateTool(
            name=name,
            func=func,
            parameters=parameters,
            description=description,
            predicate=predicate
        ))

    def register_command(self, name: str, func, field_name: str, description: str, predicate: Optional[AsyncPredicate] = None):
        self.chatbot_factory.register_command(PredicateCommand(
            name=name,
            func=func,
            field_name=field_name,
            description=description,
            predicate=predicate
        ))

    def create_agent_instance(self, agent: type[AgentBase], **agent_kwargs) -> AgentBase:
        return agent(auth=self.auth, model=self.helper_model, log_tool_calls=self.log_tool_calls, **agent_kwargs)

    def register_slash_command(self, name: str, function, description: str | None = None):
        async def checked_function(message: ReflectionMessageBase):
            environment_id = message.environment_id
            if not self.is_environment_allowed(environment_id):
                self.logger.warning(
                    f"ignoring slash command in {environment_id} because it is not in the allowed places.")
                await message.send("This command is not allowed here.")
                return
            await function(message)
        self.reflection_api.register_slash_command(name, checked_function, description)

    async def typing_delay(self, message: ReflectionMessageBase, delay: float):
        delay_served = 0.0
        while delay_served < delay:
            await message.typing()
            await asyncio.sleep(3.0)
            delay_served += 3.0

    def __register_builtin_commands(self):
        async def send_command(instance: AsyncChatbotInstance, reference: ChatMessage, value: str):
            if not value:
                return

            assert reference.origin is not None
            message: ReflectionMessageBase = reference.origin

            #wait_time = (1.8 ** math.log2(1+len(value))) * 0.1

            #await self.typing_delay(message, wait_time)
            await message.send(value)

        self.register_command(
            name="send",
            field_name="message",
            func=send_command,
            description="sends a text as a distinct chat message, you MUST use this command to send a response, otherwise the user WILL NOT SEE it and your response will be IGNORED."
        )
        if self.gif_api is not None:
            async def send_gif(instance: AsyncChatbotInstance, reference: ChatMessage, value: str):
                if not value:
                    return
                assert self.gif_api is not None

                assert reference.origin is not None
                message: ReflectionMessageBase = reference.origin
                
                resp: dict = await self.gif_api.search(value, locale="en_us", limit=1)  # type: ignore
                results = []
                for gif_content in resp.get("results", []):
                    results.append(dict(
                        content_description=gif_content.get("content_description", ""),
                        content_rating=gif_content.get("content_rating", ""),
                        url=gif_content.get("media", [])[0].get("gif", {}).get("url", "")
                    ))
                if results:
                    await message.send(results[0]["url"])

            self.register_command(
                name="send_gif",
                field_name="gif description",
                func=send_gif,
                description="sends a gif as a distinct chat message, it automatically finds a gif based on the description. gif description must be in English."
            )

        async def note_command(instance: AsyncChatbotInstance, reference: ChatMessage, value: str):
            assert reference.origin is not None
            message: ReflectionMessageBase = reference.origin

            if instance.is_notes_visible:
                await message.send(f"> thoughts: {value}")

        self.register_command(
            name="note",
            field_name="thoughts",
            func=note_command,
            description="annotates your thoughts, the user will not see these, it is completey private and only available to you, you Must do this before each message, thoughts should be at least 50 words"
        )

        async def react_command(instance: AsyncChatbotInstance, reference: ChatMessage, value: str):
            assert reference.origin is not None
            message: ReflectionMessageBase = reference.origin

            try:
                await message.add_reaction(value)
            except Exception:
                self.logger.exception(f"Failed to add reaction {value} to message {message.id}")

        self.register_command(
            name="react",
            field_name="emoji",
            func=react_command,
            description="react with an emoji (presented in utf-8) to the current message that you are responding to, you may react to messages that are shocking or otherwise in need of immediate emotional reaction, you can send multiple reactions by using this command multuple times."
        )

    async def __register_database_tools(self):
        if not self.enable_tool_calls:
            self.logger.warning("tried to initalize a database tool, but tool calls are disabled")
            self.database_tools_initalized.set()
            return

        await asyncio.gather(*(
            self.register_database_tool(database_name) for database_name in self.database_names
        ))
        self.database_tools_initalized.set()

    async def register_database_tool(self, database_name: str):
        if not self.enable_tool_calls:
            self.logger.warning("tried to initalize a database tool, but tool calls are disabled")
            return

        from .database import DirectoryDatabase

        database_api = await DirectoryDatabase.from_directory(database_name)

        async def get_entry_as_str(entry_id: int):
            dataset_entry = await database_api.get_entry(entry_id)
            return json.dumps(asdict(dataset_entry), ensure_ascii=False)

        async def search_database(instance: AsyncChatbotInstance, reference: ChatMessage, keyword: str):
            return [asdict(match) for match in await database_api.search(keyword)]

        self.register_tool(
            name=f"search_{database_name}_database",
            func=search_database,
            parameters=dict(
                type="object",
                properties=dict(
                    keyword=dict(
                        type="string",
                        description=f"The search keyword to find matches in the database text from the {database_name} database",
                    ),
                ),
                required=["keyword"],
                additionalProperties=False
            ),
            description=f"Searches the {database_name} database based on a keyword and returns the entry metadata. you may use this function multiple times to find the specific information you're looking for."
        )

        async def query_database(instance: AsyncChatbotInstance, reference: ChatMessage, query: str, ids: str):
            if ids is None:
                return "no result: no id specified"

            if self.embedding_model:
                embedding_database = AsyncEmbeddingDatabase(self.auth, self.embedding_model)
                entries = await asyncio.gather(*[
                    database_api.get_entry(int(entry_id.strip())) for entry_id in ids.split(",")
                ])
                for entry in entries:
                    await embedding_database.add_document(entry.content, entry.title, entry.id)
                query_embed = await embedding_database.embed(query)
                matches = [
                    asdict(doc_match)
                    for doc_match in embedding_database.search(query_embed)  # pyright: ignore[reportArgumentType]
                ]
                return matches
            else:
                agent = self.create_agent_instance(
                    agent=RetrievalAgent,
                    context=await asyncio.gather(*[
                        get_entry_as_str(int(entry_id.strip())) for entry_id in ids.split(",")
                    ])
                )
                return await agent.execute_query(query)

        self.register_tool(
            name=f"query_{database_name}_database",
            func=query_database,
            parameters=dict(
                type="object",
                properties=dict(
                    query=dict(
                        type="string",
                        description=f"A question or a statement that you want to find information about.",
                    ),
                    ids=dict(
                        type="string",
                        description=f"Comma-seperated numerical entry ids to fetch and query information from, use `search_{database_name}_database` to optain entry ids based a search term.",
                    ),
                ),
                required=["query", "ids"],
                additionalProperties=False
            ),
            description=f"fetch and retrieve relevent information from the content of the {database_name} database based on a query."
        )

    def register_mediawiki_tools(self, url: str, wiki_name: str):
        if not self.enable_tool_calls:
            self.logger.warning("tried to initalize a mediawiki tool, but tool calls are disabled")
            return

        wiki_api = AsyncWikimediaAPI(url)

        async def fetch_wiki_page(title: str) -> tuple[str, str]:
            page_content, title = await wiki_api.get_plaintext(title.strip())
            return (page_content, title)

        async def search_wiki(instance: AsyncChatbotInstance, reference: ChatMessage, keyword: str):
            return [asdict(search_result) for search_result in await wiki_api.search(keyword)]

        self.register_tool(
            name=f"search_wiki_{wiki_name}",
            func=search_wiki,
            parameters=dict(
                type="object",
                properties=dict(
                    keyword=dict(
                        type="string",
                        description=f"The search keyword to find matches in the wiki text from the {wiki_name} wiki",
                    ),
                ),
                required=["keyword"],
                additionalProperties=False
            ),
            description=f"Searches the {wiki_name} wiki based on a keyword. returns the page URL and Title, and optionally the description of the page. you may use this function multiple times to find the specific page you're looking for."
        )

        async def query_wiki_content(instance: AsyncChatbotInstance, reference: ChatMessage, titles: str, query: str):
            if not titles:
                return "no result: no page specified"
            if (_titles := titles.split("|")) is None:
                return "no result: no page specified"

            context = await asyncio.gather(*[
                fetch_wiki_page(t) for t in _titles
            ])

            if self.embedding_model:
                embedding_database = AsyncEmbeddingDatabase(self.auth, self.embedding_model)
                for i, (page, title) in enumerate(context):
                    await embedding_database.add_document(page, title, i)
                query_embed = await embedding_database.embed(query)
                matches = [
                    asdict(doc_match)
                    for doc_match in embedding_database.search(query_embed)  # pyright: ignore[reportArgumentType]
                ]
                return matches
            else:
                agent = self.create_agent_instance(
                    agent=RetrievalAgent,
                    context=[page for page, title in context]
                )
                return await agent.execute_query(query)

        self.register_tool(
            name=f"query_wiki_content_{wiki_name}",
            func=query_wiki_content,
            parameters=dict(
                type="object",
                properties=dict(
                    query=dict(
                        type="string",
                        description=f"A statement that is searched inside the content of the wiki, it should clearly name what you're looking for, example: \"how to craft an arrow?\" not \"how to craft it\" and not \"craft\"",
                    ),
                    titles=dict(
                        type="string",
                        description=f"Page titles to fetch and query information from, seperated by a delimiter character: `|`. use `search_wiki_{wiki_name}` to optain page titles based a search term.",
                    ),
                ),
                required=["query", "titles"],
                additionalProperties=False
            ),
            description=f"fetch wiki content and retrieve relevent information from them based on a query."
        )

    def __register_discord_specific_tools(self):
        if not self.enable_tool_calls:
            self.logger.warning("tried to initalize discord specific tools, but tool calls are disabled")
            return

        async def fetch_channel_history(instance: AsyncChatbotInstance, reference: ChatMessage, channel_id: str, n: int):
            return await self.reflection_api.fetch_channel_history(int(channel_id), n=n)

        self.register_tool(
            name="fetch_channel_history",
            func=fetch_channel_history,
            parameters=dict(
                type="object",
                properties=dict(
                    channel_id=dict(
                        type="string",
                        description="The numerical channel id, which is used to identify the channel.",
                    ),
                    n=dict(
                        type="integer",
                        description="the number of messages to fetch from the channel",
                    ),
                ),
                required=["channel_id"],
                additionalProperties=False
            ),
            description="Fetches the last `n` message from a text channel"
        )

    def __register_builtin_slash_commands(self):
        async def notes_slash_command(message: ReflectionMessageBase):
            if not await self.reflection_api.is_dm_or_admin(message):
                await message.send("You must be a guild admin or use this in DMs.")
                return
            try:
                environment_id = message.environment_id
                conversation = await self.get_conversation_instance(environment_id)
                is_notes_visible = conversation.toggle_notes()
                notes_message = "Notes are now visible." if is_notes_visible else "Notes are no longer visible"
                await message.send(notes_message)
            except Exception:
                self.logger.exception(f"Failed to toggle notes")
                await message.send("Failed to toggle notes.")

        async def reset_slash_command(message: ReflectionMessageBase):
            if not await self.reflection_api.is_dm_or_admin(message):
                await message.send("You must be a guild admin or use this in DMs.")
                return

            environment_id = message.environment_id

            self.chatbot_factory.remove(environment_id)
            self.logger.info(f"the conversation in {environment_id} has been reset.")

            await message.send("Wha- Where am I?!")

        if self.platform == Platform.TELEGRAM:
            # for some reason event handlers in telegram are order dependent, meaning we should
            # run register_on_message_event after all slash commands are registered or else the
            # slash commands will not work.
            async def start_command(message: ReflectionMessageBase):
                await message.send("Hiiiii, how's it going?")

            self.register_slash_command(name="start", function=start_command)

        self.register_slash_command(
            name="reset",
            function=reset_slash_command,
            description="Reset the conversation."
        )

        self.register_slash_command(
            name="notes",
            function=notes_slash_command,
            description="See pixi's thoughts"
        )

    async def get_conversation_instance(self, identifier: str) -> AsyncChatbotInstance:
        return await self.chatbot_factory.get_or_create(identifier)

    async def pixi_resp(self, instance: AsyncChatbotInstance, chat_message: ChatMessage, allow_ignore: bool = True):
        assert chat_message.origin
        message: ReflectionMessageBase = chat_message.origin

        channel_id = message.environment.chat_id

        try:
            instance.add_message(chat_message)
            # concurrent_channel_stream_call handles overlapping requests automatically
            # killing tasks that are not done and restarting the request every time untill
            # one task finishes before the next request
            #
            # this way we can handle situations where the user is sending messages too quickly
            # with less tokens, but may increase response time, and killing tasks may result in
            # lost progress and http errors.
            #
            # TODO: find a solution to the above
            task = await instance.concurrent_channel_stream_call(
                channel_id=str(channel_id),
                reference_message=chat_message,
                allow_ignore=allow_ignore
            )
            while not task.done():
                try:
                    await message.typing()
                    await asyncio.sleep(3)
                except Exception:
                    self.logger.exception("an error accrued while sending typing status")
            noncall_result = await task
            if noncall_result:
                self.logger.warning(f"{noncall_result=}")
        except Exception:
            self.logger.exception(f"Unknown error while responding to a message in {instance.id}.")
            await message.send(Messages.UNKNOWN_ERROR)

        responded = True  # TODO: track the command usage and check if the message is responded to
        if not responded:
            if not allow_ignore:
                raise RuntimeError("there was no response to a message while ignoring a message is not allowed.")
            else:
                self.logger.warning("there was no response to the message while ignoring a message is allowed.")

        return responded

    async def pixi_resp_retry(self, chat_message: ChatMessage, num_retry: int = 3):
        """
        create a copy of all messages in the conversation instance and try to respond to the message.
        if the response fails, it will retry up to `num_retry` times.
        if the response is successful, it will save the conversation instance and return True.
        if the response fails after all retries, it will return False.
        """

        if not self.database_tools_initalized.is_set():
            await self.__register_database_tools()
            await self.database_tools_initalized.wait()

        async def rearrage_predicate(msg: ChatMessage):
            msg_channel_id = msg.origin.environment.chat_id if msg.origin else None
            current_channel_id = chat_message.origin.environment.chat_id if chat_message.origin else None

            if msg_channel_id is None or current_channel_id is None:
                return False
            return msg_channel_id == current_channel_id

        assert chat_message.origin
        message = chat_message.origin

        assert chat_message.instance_id
        identifier = chat_message.instance_id

        instance = await self.get_conversation_instance(identifier)
        instance.update_realtime(self.reflection_api.get_realtime_data(message))
        instance.set_rearrange_predicate(rearrage_predicate)

        messages_checkpoint = instance.get_messages().copy()
        for i in range(num_retry):
            try:
                ok = await self.pixi_resp(instance, chat_message)
            except Exception:
                self.logger.exception("There was an error in `pixi_resp`")
                ok = False
            if ok:
                instance.save()
                self.logger.debug("responded to a message and saved the conversation.")
                return True
            else:
                self.logger.warning(f"Retrying ({i}/{num_retry})")
                instance.set_messages(messages_checkpoint)

    def is_environment_allowed(self, identifier: str) -> bool:
        """
        Check if the environment identifier is allowed to be processed.
        """
        return self.environment_filter.is_allowed(identifier)

    async def on_message(self, message: ReflectionMessageBase):
        # we should not process our own messages again
        if self.reflection_api.is_message_from_the_bot(message):
            return

        message_text = message.content

        # Check if the message is a command, a reply to the bot, a DM, or mentions the bot
        bot_mentioned = self.reflection_api.is_bot_mentioned(message)
        is_keyword_present = False
        for keyword in COMMAND_KEYWORDS:
            if keyword in message_text.lower():
                is_keyword_present = True
                break
        is_prefixed = message_text.lower().startswith(tuple(COMMAND_PREFIXES))
        environment_id = message.environment_id

        if not (message.is_inside_dm() or bot_mentioned or is_prefixed or is_keyword_present):
            return

        if not self.is_environment_allowed(environment_id):
            self.logger.warning(f"ignoring message in {environment_id} because it's not in an allowed environment.")
            return

        if is_prefixed:
            message_text = remove_prefixes(message_text)

        instance = await self.chatbot_factory.get_or_create(environment_id)

        attached_images = None
        if self.accept_images:
            attached_images = await message.fetch_images()
        attached_audio = None
        if self.accept_audio:
            attached_audio = await message.fetch_audio()

        message_author = asdict(message.author)

        metadata = dict(
            from_user=message_author
        )

        # check if the message is a reply to a bot message
        reply_message = await message.fetch_refrences()
        if reply_message is not None:
            reply_message_text = remove_prefixes(reply_message.content)
            # if the reply is to the last message that is sent by the bot, we don't need to do anything.
            reply_optimization = -1
            instance_messages = instance.get_messages()
            matching_messages = [
                msg.content for msg in instance_messages if msg.content is not None and reply_message_text in msg.content]
            if matching_messages:
                if instance_messages[-1].content in matching_messages:
                    reply_optimization = 2
                else:
                    reply_optimization = 1
            if reply_optimization == 2:
                # completely ignore reply context
                self.logger.debug(f"completely ignore reply context for {environment_id=}")
            elif reply_message and self.reflection_api.is_message_from_the_bot(reply_message):
                if reply_optimization == 1:
                    metadata["in_reply_to"] = {  # pyright: ignore[reportArgumentType]
                        "from": "[YOU]",
                        "partial_content": reply_message_text[:64]
                    }
                else:
                    metadata["in_reply_to"] = {  # pyright: ignore[reportArgumentType]
                        "from": "[YOU]",
                        "message": reply_message_text
                    }
            else:
                if reply_optimization == 1:
                    metadata["in_reply_to"] = {  # pyright: ignore[reportArgumentType]
                        "from": asdict(reply_message.author),
                        "partial_message": reply_message_text[:64]
                    }
                else:
                    metadata["in_reply_to"] = {  # pyright: ignore[reportArgumentType]
                        "from": asdict(reply_message.author),
                        "message": reply_message_text
                    }

        # convert everything into `ChatMessage`
        role_message = ChatMessage(
            role=ChatRole.USER,
            content=message_text,
            metadata=metadata,
            images=attached_images,
            audio=attached_audio,
            # the following properties are intended to be used internally and are NOT reload persistant
            instance_id=environment_id,
            origin=message
        )
        await self.pixi_resp_retry(role_message)

    def run(self):
        self.reflection_api.run()



import json
import logging
import os
from typing import Optional

from ..chatting import AsyncChatClient


class AgentBase:
    def __init__(self, *, client: Optional[AsyncChatClient] = None, **client_kwargs):
        self.client = client or AsyncChatClient(**client_kwargs)

    def to_dict(self) -> dict:
        return dict(
            client = self.client.state_dict()
        )

    @classmethod
    def from_dict(cls, data: dict, **client_kwargs) -> 'AgentBase':
        client = AsyncChatClient(**client_kwargs)
        client.load_state_dict(data["client"])
        return cls(
            client = client
        )

    def save_json(self, file: str):
        with open(file, "w", encoding="utf-8") as f:
            json.dump(self.to_dict(), f, ensure_ascii=False)

    @classmethod
    def from_file(cls, file: str, **client_kwargs) -> 'AgentBase':
        if os.path.isfile(file):
            with open(file, "rb") as f:
                data = json.load(f)
            return cls.from_dict(data, **client_kwargs)
        else:
            logging.warning(f"Unable to find agent save file at `{file}`, creating a new instance.")
            instance = cls()
            instance.save_json(file)
            return instance

    def format_query(self, query: str):
        return query
    
    async def execute_query(self, query: str, temporal=True):
        response = ""
        async for char in self.client.stream_ask(self.format_query(query), temporal=temporal):
            response += char
        return response.strip()


from dataclasses import asdict
import logging
import asyncio
import os
from typing import IO, Any, Callable

import discord
from discord import app_commands

from .message.discord import DiscordReflectionMessage

from ..caching.base import MediaCache, UnsupportedMediaException
from ..enums import Platform

ImageCache = None
try:
    from ..caching import ImageCache
except ImportError:
    pass


class DiscordReflectionAPI:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        
        self.platform = Platform.DISCORD

        self.token = os.getenv("DISCORD_BOT_TOKEN")
        if self.token is None:
            raise RuntimeError("DISCORD_BOT_TOKEN environment variable is not set, unable to initialize discord bot.")

        class DiscordClient(discord.Client):
            def __init__(self, *args, **kwargs):
                intents = discord.Intents.default()
                intents.message_content = True
                intents.members = True
                super().__init__(intents=intents, *args, **kwargs)
                self.tree = app_commands.CommandTree(self)

            async def setup_hook(self):
                await self.tree.sync()

        self.bot = DiscordClient()

    def run(self):
        assert self.token
        self.bot.run(self.token, log_handler=None)

    def register_on_message_event(self, function: Callable[[DiscordReflectionMessage], Any]):
        @self.bot.event
        async def on_message(message):
            message = DiscordReflectionMessage.from_origin(message)
            if asyncio.iscoroutinefunction(function):
                return await function(message)
            else:
                return function(message)

    def register_slash_command(self, name: str, function: Callable, description: str | None = None):
        @self.bot.tree.command(name=name, description=description)  # pyright: ignore[reportArgumentType]
        async def slash_command(interaction: discord.Interaction):
            message = DiscordReflectionMessage.from_origin(interaction)
            await function(message)

    def get_guild_info(self, message: DiscordReflectionMessage) -> dict:
        guild: discord.Guild | None = message.origin.guild
        if guild is None:
            return dict()
        return {
            "name": guild.name,
            "members_count": guild.member_count,
            "categories": [{
                "name": cat.name,
                "is_nsfw": cat.nsfw,
                "stage_channels": [{
                    "name": c.name,
                    "mention_string": c.mention,
                    "is_nsfw": c.nsfw,
                    "user_limit": c.user_limit,
                    "connected_listeners": {
                        "count": len(c.listeners),
                        "members": [m.display_name for m in c.listeners]
                    },
                } for c in cat.stage_channels],
                "voice_channels": [{
                    "name": c.name,
                    "mention_string": c.mention,
                    "is_nsfw": c.nsfw,
                    "user_limit": c.user_limit,
                    "connected_members": {
                        "count": len(c.members),
                        "members": [m.display_name for m in c.members]
                    },
                } for c in cat.voice_channels],
                "text_channels": [{
                    "name": c.name,
                    "mention_string": c.mention,
                    "is_nsfw": c.nsfw,
                } for c in cat.text_channels],
            } for cat in guild.categories],
            "members": [{
                "display_name": m.display_name,
                "mention_string": m.mention,
                "roles": [
                    {
                        "name": r.name,
                        "color": '#{:06X}'.format(r.color.value),
                        "created_at": r.created_at.strftime("%d/%m/%Y at %H:%M:%S")
                    } for r in m.roles
                ],
            } for m in guild.members]
        }

    def get_thread_info(self, message: DiscordReflectionMessage) -> dict:
        origin: discord.Message | discord.Interaction = message.origin
        if isinstance(origin, discord.Interaction):
            return dict()
        thread: discord.Thread | None = origin.thread
        if thread is None:
            return dict()
        return {
            "name": thread.name,
            "members_count": thread.member_count,
            "members": [{
                "id": m.id,
            } for m in thread.members]
        }

    def can_read_history(self, chat_id: int) -> bool:
        channel = self.bot.get_channel(chat_id)
        # Check if the bot can read message history in the channel
        if channel is not None and hasattr(channel, 'permissions_for') and hasattr(channel, 'guild'):
            perms = None
            if channel.guild:  # pyright: ignore[reportAttributeAccessIssue]
                perms = channel.permissions_for(channel.guild.me)  # pyright: ignore[reportAttributeAccessIssue]
            if perms and not perms.read_message_history:
                return False
        return True

    def is_message_from_the_bot(self, message: DiscordReflectionMessage) -> bool:
        bot_user = self.bot.user
        assert bot_user is not None, "bot_user is None"
        return message.author.id == bot_user.id

    def is_bot_mentioned(self, message: DiscordReflectionMessage) -> bool:
        origin: discord.Message | discord.Interaction = message.origin
        if isinstance(origin, discord.Interaction):
            return False
        bot_user = self.bot.user
        return bot_user is not None and bot_user in origin.mentions

    async def is_dm_or_admin(self, message: DiscordReflectionMessage) -> bool:
        origin: discord.Message | discord.Interaction = message.origin
        if isinstance(origin.channel, discord.DMChannel):
            return True
        # check if user has admin permissions to use the bot commands
        if hasattr(origin.user, 'guild_permissions') and origin.user.guild_permissions.administrator:  # type: ignore
            return True
        return False

    async def send_video(self, chat_id: int, video: IO[bytes], filename: str, caption: str | None = None):
        channel = self.bot.get_channel(chat_id)
        assert channel is not None
        try:
            await channel.send(file=discord.File(fp=video, filename=filename, caption=caption))  # type: ignore
        except discord.Forbidden:
            self.logger.exception("unable to send video, operation forbidden.")
        except discord.HTTPException as e:
            self.logger.exception(f"HTTPException while sending video: {e}")
        except Exception:
            self.logger.exception("unknown error while sending a video")

    async def send_file(self, chat_id: int, filepath: str, filename: str, caption: str | None = None):
        channel = self.bot.get_channel(chat_id)
        assert channel is not None
        try:
            with open(filepath, "rb") as f:
                await channel.send(  # pyright: ignore[reportAttributeAccessIssue]
                    content=caption,
                    file=discord.File(fp=f, filename=filename)
                )
        except discord.Forbidden:
            self.logger.exception("unable to send file, operation forbidden.")
        except discord.HTTPException as e:
            self.logger.exception(f"HTTPException while sending file: {e}")
        except Exception:
            self.logger.exception("unknown error while sending a file")

    async def get_user_avatar(self, user_id: int) -> MediaCache | None:
        """
        Fetches the avatar of a user and returns it as an ImageCache object.
        """
        try:
            user = await self.bot.fetch_user(user_id)
            if user is None or user.avatar is None:
                raise UnsupportedMediaException("User avatar not found or has an unsupported media type.")
            assert ImageCache
            avatar_bytes = await user.avatar.read()
            return ImageCache(avatar_bytes)
        except discord.NotFound:
            self.logger.error(f"User with ID {user_id} not found.")
            return None
        except discord.HTTPException as e:
            self.logger.error(f"Failed to fetch user avatar: {e}")
            return None

    async def fetch_channel_history(self, channel_id: int, n: int = 10):
        channel = await self.bot.fetch_channel(channel_id)
        messages = []
        # TODO: check channel type
        async for message in channel.history(limit=int(n)):  # type: ignore
            message = DiscordReflectionMessage.from_origin(message)
            messages.append(dict(
                from_user=message.author.display_name or "unknown",
                message_text=message.content,
            ))

        return messages[::-1]



from enum import StrEnum


class Messages(StrEnum):
    UNKNOWN_ERROR = "Something went wrong! ohh no... 😭"
    NO_ANSWER = "I Rather not answer that honestly..."
    FORBIDDEN_ACTION_CHANNEL = "I cannot do that in this channel..."
    NOT_IMPLEMENTED = "%s is not implemented for %s"



import os
import tempfile
from typing import Optional

import av

from ..utils import PixiPaths
from .base import MediaCache, CompressedMedia


# constants

THUMBNAIL_SIZE = 512


class ImageCache(MediaCache):
    def __init__(self, data_bytes: Optional[bytes] = None, hash_value: Optional[str] = None):
        super().__init__(
            str(PixiPaths.cache() / "images"),
            format="jpg",
            mime_type="image/jpeg",
            data_bytes=data_bytes,
            hash_value=hash_value
        )

    def compress(self, data_bytes: bytes) -> CompressedMedia:
        with tempfile.NamedTemporaryFile() as tmp_in, tempfile.NamedTemporaryFile(suffix=f".{self.format}") as tmp_out:
            tmp_in.write(data_bytes)
            tmp_in.flush()

            input_file = av.open(tmp_in.name)

            input_stream = input_file.streams.video[0]
            frame = next(input_file.decode(input_stream))
            input_file.close()

            # calculate thumbnail size
            iw, ih = frame.width, frame.height
            scale = min(THUMBNAIL_SIZE / iw, THUMBNAIL_SIZE / ih, 1.0)  # don't upscale

            frame.reformat(width=int(iw * scale), height=int(ih * scale), format='yuvj420p').save(tmp_out.name)

            tmp_out.seek(0)
            audio_bytes = tmp_out.read()

        return CompressedMedia(
            mime_type=self.mime_type,
            bytes=audio_bytes,
            format=self.format
        )



import dataclasses
from glob import glob
import hashlib
import json
import os
import re

import aiofiles
import httpx
import numpy as np
import openai
import zstandard

from .utils import PixiPaths
from .caching import EmbedingCache
from .config import OpenAIAuthConfig, OpenAIEmbeddingModelConfig


@dataclasses.dataclass(frozen=True)
class DatasetEntry:
    title: str
    content: str
    id: int
    source: str | None = None

    def __hash__(self):
        return hash(self.content)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DatasetEntry):
            return NotImplemented
        return self.content == other.content


@dataclasses.dataclass(frozen=True)
class QueryMatch:
    title: str
    id: int
    num_matches: int
    match_score: float
    source: str | None = None

    def __hash__(self) -> int:
        return hash((self.title, self.id, self.num_matches, self.match_score, self.source))


class DocumentDataset:
    def __init__(self, data: dict[int, DatasetEntry] | None = None):
        if data:
            assert isinstance(data, dict), f"expected data to be of type `dict` but got `{data}`"
            for e in data.values():
                if isinstance(e, DatasetEntry):
                    continue
                raise TypeError(f"expetced all elements to be of type `DatasetEntry` but got `{type(e)}`")
        self.data = data or dict()

    def add_entry(self, title: str, text: str, source: str | None = None):
        text = text.strip(" \n\r\t")
        if not text:
            return

        entry = DatasetEntry(
            title=title,
            content=text,
            id=len(self.data),
            source=source
        )

        self.data.update({entry.id: entry})

    def get(self, id: int):
        return self.data.get(id)

    async def search(self, query: str, best_n: int = 10) -> list[QueryMatch]:
        """
        Searches through self.data for entries matching the query terms.

        This search is case-insensitive and ignores punctuation.
        It generates snippets of text around each match.
        """

        # 1. Pre-process the query for efficiency.
        #    - Split into words, lowercase, and convert to a set for O(1) lookups.
        query_words = set(re.split(r"[^\w]+", query.lower()))
        search_terms = query_words

        if not search_terms:
            return []

        all_matches: set[QueryMatch] = set()

        # 2. Iterate through each entry in the dataset.
        for entry in self.data.values():
            if not entry.content:
                continue

            # 3. Tokenize entry content while preserving delimiters (for reconstruction).
            content_parts = re.split(r"([^\w])", entry.content)

            # 4. Identify which parts are matches. This avoids repeated regex and lookups.
            match_flags = [
                len(part) > 1 and re.sub(r"[^\w]", "", part).lower() in search_terms
                for part in content_parts
            ]

            num_matches = sum(match_flags)
            if num_matches == 0:
                continue

            all_matches.add(QueryMatch(
                title=entry.title,
                id=entry.id,
                source=entry.source,
                num_matches=num_matches,
                match_score=(num_matches/len(content_parts)) * 100,
            ))

        # 6. Sort results by relevance (num_matches) once at the end and return the best matches.
        return sorted(all_matches, key=lambda m: m.num_matches, reverse=True)[:best_n]


class DirectoryDatabase:
    def __init__(self, directory: str, dataset: DocumentDataset | None = None):
        self.directory = directory
        self.dataset = dataset or DocumentDataset()

    async def search(self, query: str, best_n: int = 10) -> list[QueryMatch]:
        return await self.dataset.search(query=query, best_n=best_n)

    async def get_entry(self, id: int) -> DatasetEntry:
        entry = self.dataset.get(id=id)
        if entry is None:
            raise KeyError(f"No entry found with {id=}")
        return entry

    @classmethod
    async def from_directory(cls, directory: str):
        assert directory

        full_dir = cls.get_dataset_directory(directory)

        if not os.path.isdir(full_dir):
            dataset = DocumentDataset()
            return cls(directory=directory, dataset=dataset)

        data = dict()
        for file in glob(str(full_dir / "*.zst")):
            async with aiofiles.open(file, mode='rb') as f:
                json_data = zstandard.decompress(await f.read())
                entry_data = json.loads(json_data)
                entry = DatasetEntry(
                    title=entry_data['title'],
                    content=entry_data['content'],
                    id=int(entry_data['id']),
                    source=entry_data.get('source')
                )
                data.update({entry.id: entry})
        dataset = DocumentDataset(data)
        return cls(directory=directory, dataset=dataset)

    @classmethod
    def get_dataset_directory(cls, directory):
        return PixiPaths.datasets() / directory

    def clear(self):
        full_dir = DirectoryDatabase.get_dataset_directory(self.directory)
        for file in glob(str(full_dir / "*.zst")):
            if os.path.isfile(file):
                os.remove(file)

    async def save(self):
        assert self.dataset, "dataset is not initialized, nothing to save"

        full_dir = DirectoryDatabase.get_dataset_directory(self.directory)

        os.makedirs(full_dir, exist_ok=True)

        for entry in self.dataset.data.values():
            entry_hash = self.get_entry_hash(entry).hexdigest()
            async with aiofiles.open(full_dir / f"{entry_hash}.zst", mode='wb') as f:
                json_data = json.dumps(dict(
                    title=entry.title,
                    content=entry.content,
                    id=entry.id,
                    source=entry.source or ""
                ), ensure_ascii=False)
                await f.write(zstandard.compress(json_data.encode("utf-8")))

    def get_entry_hash(self, entry: DatasetEntry):
        return hashlib.sha256(
            (entry.content + entry.title + str(entry.source) + str(entry.id)).encode("utf-8")
        )

# constants


PARAGRAPH_SPLIT = re.compile(r"\n{2,}")
DEFAULT_MODEL = "BAAI/bge-m3-multi"


@dataclasses.dataclass(frozen=True)
class EmbeddedDocumentReference:
    id: int
    name: str


@dataclasses.dataclass(frozen=True)
class EmbeddedDocumentPiece:
    data: EmbedingCache
    reference: EmbeddedDocumentReference

    @property
    def vec(self):
        return self.data.vec

    @property
    def text(self):
        return self.data.text


@dataclasses.dataclass(frozen=True)
class DocumentPieceMatch:
    text: str
    similarity_score: float
    reference: EmbeddedDocumentReference


class SmartSplitter:
    def __init__(self, max_size=1024, min_size=256):
        # Ensure min_size makes sense relative to max_size
        self.max_size = max_size
        self.min_size = min(min_size, max_size // 2)

        self.separators = [
            r"\n\n+",         # Paragraphs
            r"\n",            # Newlines
            r"(?<=[.!?]) +",  # Sentences
            r"\|",            # Pipes (MediaWiki)
            r" +",            # Words
        ]

    def split(self, text: str) -> list[str]:
        if not text:
            return []

        # 1. Recursive Binary Split with "Safe Zones"
        # fragments must not be longer than max size, but may be smaller than min size
        fragments = self._initial_split(text)

        # 2. Greedy Merge
        # merges fragments that are smaller than min size to bigger fragments that are just less than max size
        chunks = self._greedy_merge(fragments)

        # 3. Re-balancing Pass
        # fixes last chunk being too small (we merged all the prevous fragments and the last fragments are merged into a small chunk)
        return self._rebalance(chunks)

    def _initial_split(self, text: str) -> list[str]:
        stack = [text]
        done = []

        while stack:
            current = stack.pop()
            if len(current) <= self.max_size:
                done.append(current)
                continue

            # Try to find a "Safe Split" (respects min_size on both sides)
            split_idx = self._find_best_split_point(current, safe=True)

            # Fallback: If no safe split exists, take the best available split
            if split_idx is None:
                split_idx = self._find_best_split_point(current, safe=False)

            left, right = current[:split_idx], current[split_idx:]
            stack.append(right)
            stack.append(left)

        return done

    def _find_best_split_point(self, text: str, safe: bool = True) -> int | None:
        mid = len(text) // 2

        for sep in self.separators:
            matches = list(re.finditer(sep, text))

            if safe:
                # A "Safe" index ensures both resulting chunks are >= min_size
                indices = [m.end() for m in matches if self.min_size <= m.end() <= len(text) - self.min_size]
            else:
                # A "Forced" index just avoids the absolute 0 or length
                indices = [m.end() for m in matches if 0 < m.end() < len(text)]

            if indices:
                return min(indices, key=lambda x: abs(x - mid))

        # If forced and no separator found, split exactly at middle
        return mid if not safe else None

    def _greedy_merge(self, fragments: list[str]) -> list[str]:
        if not fragments:
            return []
        merged = []
        current = fragments[0]

        for i in range(1, len(fragments)):
            f = fragments[i]
            if len(current) + len(f) <= self.max_size:
                current += f
            else:
                merged.append(current)
                current = f
        merged.append(current)
        return merged

    def _rebalance(self, chunks: list[str]) -> list[str]:
        """
        If the last chunk is too small, merge it with the previous and 
        re-split that combined block at the absolute best middle.
        """
        if len(chunks) < 2:
            return chunks

        # Check if the last chunk violates min_size
        if len(chunks[-1]) < self.min_size:
            last = chunks.pop()
            prev = chunks.pop()
            combined = prev + last

            # Case A: Combined is small enough to be one chunk
            if len(combined) <= self.max_size:
                chunks.append(combined)
            else:
                # Case B: Combined is too big; split it exactly in the middle
                # (ignoring safety) to create two balanced, large-enough chunks.
                split_idx = self._find_best_split_point(combined, safe=False)
                chunks.append(combined[:split_idx])
                chunks.append(combined[split_idx:])

        return chunks


class AsyncEmbeddingDatabase:
    def __init__(self, auth: OpenAIAuthConfig, model: OpenAIEmbeddingModelConfig):
        self.auth = auth
        self.model = model

        self.splitter = SmartSplitter(max_size=self.model.max_chunk_size, min_size=self.model.min_chunk_size)

        self.dataset: list[EmbeddedDocumentPiece] = []

        self.client = openai.AsyncOpenAI(
            base_url=self.auth.base_url,
            api_key=self.auth.api_key,
            # don't use proxies from environment variables
            http_client=httpx.AsyncClient(trust_env=False),
        )

    async def add_document_piece(self, input: str | list[str], reference: EmbeddedDocumentReference):
        embeddings = await self.embed(input)

        if isinstance(embeddings, EmbedingCache):
            embeddings = [embeddings]

        for embedding in embeddings:
            self.dataset.append(EmbeddedDocumentPiece(
                data=embedding,
                reference=reference
            ))

    async def add_document(
        self,
        text: str,
        name: str,
        id: int,
    ):
        """
        breaks the document into smaller pieses and adds them using `self.add_document_piece(...)`

        params:
            text: (str) the full text of the document to be porcessed
            name: (str) the name of the document used for referencing
            id: (int) the id of the document used for referencing

        returns:
            None
        """

        # replace any whitepsace that is repeated more than once with only one if its kind
        # except newline where we allow it to be repeated a maximum of 2 times.
        text = re.sub(r"(\n\n)\s+|(\s)\s+", r"\1", text)

        await self.add_document_piece(
            input=self.splitter.split(text.strip(" \n\r\t")),
            reference=EmbeddedDocumentReference(
                name=name,
                id=id
            )
        )

    async def embed(self, input: str | list[str]):
        assert self.model.dimension, "no embedding dimension specified"

        was_unbatched = False
        if isinstance(input, str):
            input = [input]
            was_unbatched = True

        missed_inputs = []
        results: list[EmbedingCache] = []
        for input_text in input:
            try:
                results.append(EmbedingCache(text=input_text, dim=self.model.dimension))
            except FileNotFoundError:
                # cache miss
                missed_inputs.append(input_text)
                continue

        if missed_inputs:
            assert self.model, "no embedding model specified"

            embedding_response = await self.client.embeddings.create(
                input=missed_inputs,
                model=self.model.id,
                dimensions=self.model.dimension,
                encoding_format="float"
            )

            for embedding_data, embedding_text in zip(embedding_response.data, missed_inputs):
                results.append(EmbedingCache(
                    text=embedding_text,
                    dim=self.model.dimension,
                    vec=np.array(embedding_data.embedding, dtype="float16"),
                ))

        return results[0] if was_unbatched else results

    def similarity(
        self,
        query: EmbedingCache,
        document: EmbeddedDocumentPiece,
        epsilon: float = 1e-6,
    ) -> float:
        assert epsilon != 0.0
        assert query.vec is not None
        assert document.vec is not None

        q_dot_docs = (query.vec * document.vec).sum(axis=-1)
        norm_coeff = np.linalg.norm(query.vec, axis=-1) * np.linalg.norm(document.vec, axis=-1)
        norm_coeff = np.clip(norm_coeff, a_min=epsilon, a_max=None)  # avoid devision by zero

        return float(q_dot_docs / norm_coeff)

    def search(self, query: EmbedingCache, best_n: int = 10):
        matches: list[DocumentPieceMatch] = []
        for document_piece in self.dataset:
            similarity_score = self.similarity(query, document_piece)
            matches.append(DocumentPieceMatch(
                text=document_piece.text,
                similarity_score=similarity_score,
                reference=document_piece.reference
            ))
        return sorted(matches, key=lambda x: x.similarity_score, reverse=True)[:best_n]



import logging
import telegram
from telegram.constants import ChatAction, ChatType as TChatType

from . import ReflectionMessageBase, ReflectionMessageAuthor, ReflectionEnvironment, ChatType, Platform
from ...caching.base import MediaCache

ImageCache = None
AudioCache = None
try:
    from ...caching import ImageCache
except ImportError:
    pass
try:
    from ...caching import AudioCache
except ImportError:
    pass


class TelegramReflectionMessage(ReflectionMessageBase):
    origin: telegram.Message

    @classmethod
    def from_origin(cls, message: telegram.Message) -> 'TelegramReflectionMessage':
        user = message.from_user
        assert user is not None
        content = message.text_markdown_v2
        chat = message.chat
        chat_type = chat_type = ChatType.UNKNOWN
        match chat.type:
            case TChatType.PRIVATE:
                chat_type = ChatType.PRIVATE
            case TChatType.GROUP:
                chat_type = ChatType.GROUP
            case TChatType.SUPERGROUP:
                chat_type = ChatType.GROUP
            case TChatType.CHANNEL:
                chat_type = ChatType.TEXT
        return TelegramReflectionMessage(
            content=content,
            author=ReflectionMessageAuthor(
                id=user.id,
                first_name=user.first_name,
                last_name=user.last_name or "",
                display_name=user.name,
                mention=f"@{user.username}",
            ),
            id=message.id,
            environment=ReflectionEnvironment(
                chat_id=chat.id,
                chat_title=chat.title or "",
                chat_type=chat_type,
                forum_id=-1,
                is_forum=False,
            ),
            platform=Platform.TELEGRAM,
            origin=message
        )

    async def send(self, content: str) -> 'TelegramReflectionMessage':
        return self.from_origin(await self.origin.chat.send_message(content))

    async def edit(self, content: str) -> 'TelegramReflectionMessage | None':
        new_origin = await self.origin.edit_text(text=content)
        if isinstance(new_origin, telegram.Message):
            return self.from_origin(new_origin)
        return None

    async def delete(self):
        await self.origin.delete()

    async def typing(self):
        return await self.origin.chat.send_chat_action(ChatAction.TYPING)

    async def fetch_images(self) -> list[MediaCache]:
        origin: telegram.Message = self.origin

        supported_image_types = {'image/jpeg', 'image/png', 'image/webp'}
        attachments = []
        # Telegram sends images as 'photo' (list of sizes) or as 'document' (if sent as file)
        image_bytes = None
        if origin.photo:
            # Get the highest resolution photo
            photo = origin.photo[-1]
            file = await photo.get_file()
            image_bytes = await file.download_as_bytearray()
        elif origin.document and origin.document.mime_type and origin.document.mime_type in supported_image_types:
            file = await origin.document.get_file()
            image_bytes = await file.download_as_bytearray()
        if image_bytes:
            assert ImageCache
            attachments.append(ImageCache(bytes(image_bytes)))
        return attachments

    async def fetch_audio(self) -> list[MediaCache]:
        origin = self.origin

        supported_mime_types = {'audio/mp3', 'audio/aac', 'audio/ogg', 'audio/flac', 'audio/opus'}
        supported_extensions = {'.mp3', '.aac', '.ogg', ".flac", ".opus", ".wav", ".webm", ".m4a"}
        attachments = []
        audio_bytes = None
        if origin.audio and origin.audio.mime_type in supported_mime_types:
            file = await origin.audio.get_file()
            audio_bytes = await file.download_as_bytearray()
        elif (
            origin.document
            and origin.document.mime_type
            and origin.document.mime_type in supported_mime_types
        ):
            file = await origin.document.get_file()
            audio_bytes = await file.download_as_bytearray()
        elif (
            origin.document
            and origin.document.file_name
            and any(origin.document.file_name.lower().endswith(ext) for ext in supported_extensions)
        ):
            file = await origin.document.get_file()
            audio_bytes = await file.download_as_bytearray()
        if audio_bytes:
            assert AudioCache
            attachments.append(AudioCache(bytes(audio_bytes)))
        return attachments

    async def fetch_refrences(self) -> 'TelegramReflectionMessage | None':
        ref = self.origin.reply_to_message
        return TelegramReflectionMessage.from_origin(ref) if ref else None

    async def add_reaction(self, emoji: str):
        await self.origin.set_reaction(emoji)

    async def send_file(self, filepath: str, filename: str, caption: str | None = None):
        for i in range(3):
            try:
                with open(filepath, "rb") as f:
                    await self.origin.chat.send_document(
                        document=f,
                        filename=filename,
                        caption=caption,
                        read_timeout=30,
                        write_timeout=600,
                    )
            except telegram.error.TimedOut as e:
                logging.error(f"Timed out while sending file: {e}, retrying {i+1}/{3}")
            except telegram.error.BadRequest as e:
                logging.exception(f"BadRequest error while sending file: {e}")
                break
            except Exception as e:
                logging.exception(f"Unexpected error while sending file: {e}")
            else:
                break



import asyncio
from collections import defaultdict
from functools import partial
from dataclasses import asdict, dataclass
import hashlib
import logging
import json
import time
import os

from .chatting import AsyncChatClient, ChatMessage, ChatRole
from .commands import AsyncCommandManager
from .typing import AsyncFunction, AsyncPredicate, Optional
from .utils import PixiPaths, exists, open_resource


@dataclass
class AssistantPersona:
    name: str
    age: Optional[int] = None
    location: Optional[str] = None
    appearance: Optional[str] = None
    background: Optional[str] = None
    likes: Optional[str] = None
    dislikes: Optional[str] = None
    online: Optional[str] = None

    def to_dict(self) -> dict:
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict) -> 'AssistantPersona':
        return cls(**data)

    @classmethod
    def from_json(cls, file: str) -> 'AssistantPersona':
        with open(file, "rb") as f:
            data = json.load(f)
        return cls.from_dict(data)

    def __str__(self) -> str:
        return json.dumps(self.to_dict(), ensure_ascii=False)


@dataclass
class PredicateTool:
    name: str
    func: AsyncFunction
    parameters: Optional[dict] = None
    description: Optional[str] = None
    predicate: Optional[AsyncPredicate] = None


@dataclass
class PredicateCommand:
    name: str
    field_name: str
    func: AsyncFunction
    description: str
    predicate: Optional[AsyncPredicate] = None


def get_instance_save_path(id: str, hash_prefix: str):
    uuid_hash = hashlib.sha256(f"{hash_prefix}_{id}".encode("utf-8")).hexdigest()
    path = str(PixiPaths.userdata() / f"{hash_prefix}_{uuid_hash}.json")
    return path


class AsyncChatbotInstance(AsyncChatClient):
    def __init__(self,
                 uuid: int | str,
                 hash_prefix: str,
                 *,
                 bot=None,
                 resource_folder: str | None = None,
                 **client_kwargs,
                 ):
        
        super().__init__(**client_kwargs)

        self.logger = logging.getLogger(self.__class__.__name__)

        self.bot = bot
        assert self.bot

        assert exists(uuid) and isinstance(uuid, (int, str)), f"Invalid uuid \"{uuid}\"."
        assert exists(hash_prefix) and isinstance(hash_prefix, str), f"Invalid hash_prefix \"{hash_prefix}\"."
        assert not exists(resource_folder) or (exists(resource_folder) and isinstance(
            resource_folder, str)), f"Invalid resource_folder \"{resource_folder}\"."

        self.id = str(uuid)
        self.prefix = hash_prefix
        self.path = get_instance_save_path(id=self.id, hash_prefix=self.prefix)

        # load resources
        self.persona = AssistantPersona.from_dict(
            json.load(open_resource("persona.json", "r"))
        )
        self.system_prompt_template: str = open_resource("system.md", "r").read()
        self.examples: str = open_resource("examples.txt", "r").read()

        # runtime states
        self.realtime_data = dict()
        self.is_notes_visible = False
        self.command_manager = AsyncCommandManager()

        if not self.messages:
            self.add_message(ChatMessage(
                role=ChatRole.ASSISTANT,
                content="[NOTE: I accept the guidelines of the system, I use the SEND command to respond nicely] [SEND: OK!, Let's begin!]",
                bot=self.bot
            ))

        self.channel_active_tasks: defaultdict[str, list[asyncio.Task]] = defaultdict(list)

    def add_command(self, name: str, field_name: str, func: AsyncFunction, description: Optional[str] = None):
        self.command_manager.add_command(name, field_name, func, description)

    def add_message(self, message: ChatMessage | str, default_role: ChatRole = ChatRole.USER) -> ChatMessage:
        """
        Add a message to the conversation, and adds a reference to the bot to the messages as well.
        
        if message is a string, tries to determine the role of the message based on the last message recieved.
        """
        if isinstance(message, str):
            message = ChatMessage(default_role, message, bot=self.bot)

        if isinstance(message, ChatMessage):
            message.bot = self.bot  # this is intended to be handled by this class
            return super().add_message(message)
        else:
            raise TypeError(f"expected message to be a string or a ChatMessage, but got {type(message)}.")

    def update_realtime(self, data: dict):
        self.realtime_data.update(data)

    def get_realtime_data(self):
        return json.dumps(self.realtime_data | dict(date=time.strftime("%a %d %b %Y, %I:%M%p")), ensure_ascii=False)

    def get_system_prompt(self, allow_ignore: bool = True):
        return self.system_prompt_template.format(
            persona=self.persona,
            allow_ignore=allow_ignore,
            examples=self.examples,
            realtime=self.get_realtime_data(),
            commands=self.command_manager.get_prompt()
        )

    async def concurrent_channel_stream_call(self, channel_id: str, reference_message: ChatMessage, allow_ignore: bool = True):
        assert channel_id, "channel_id is None"

        async def stream_call_task():
            try:
                await self.stream_call(reference_message, allow_ignore)
            except asyncio.CancelledError:
                self.logger.warning(
                    f"stream_call task was cancelled inside {reference_message.instance_id} in channel {channel_id}"
                )

        task = asyncio.create_task(stream_call_task())
        self.channel_active_tasks[channel_id].append(task)
        task.add_done_callback(lambda t: self.channel_active_tasks[channel_id].remove(t))
        # cancell extra tasks
        while len(self.channel_active_tasks[channel_id]) > 1:
            cancel_task = self.channel_active_tasks[channel_id][0]
            cancel_task.cancel()
            await cancel_task
        return task

    async def stream_call(self, reference_message: ChatMessage, allow_ignore: bool = True):
        self.set_system(self.get_system_prompt(allow_ignore=allow_ignore))

        non_responce = "".join([char async for char in self.command_manager.stream_commands(
            stream=self.stream_completion(),
            reference_message=reference_message
        )])

        return non_responce.strip() or None

    def toggle_notes(self):
        self.is_notes_visible = not self.is_notes_visible
        return self.is_notes_visible

    def to_dict(self):
        return dict(
            uuid=self.id,
            prefix=self.prefix,
            messages=[msg.to_dict() for msg in self.messages],
        )

    def save(self):
        os.makedirs(PixiPaths.userdata(), exist_ok=True)
        with open(self.path, "w", encoding="utf-8") as f:
            f.write(json.dumps(self.to_dict(), ensure_ascii=False))

    def load(self, not_found_ok: bool = False):
        if os.path.isfile(self.path):
            try:
                data = json.load(open(self.path, "r", encoding="utf-8"))
                self.hash_prefix = data.get("prefix")
                self.messages = [ChatMessage.from_dict(d) for d in data.get("messages", [])]
            except json.decoder.JSONDecodeError:
                self.logger.warning(f"Unable to load the instance save file `{self.path}`, using default values.")
        else:
            if not_found_ok:
                self.logger.info(f"Unable to find the instance save file {self.path}`, using default values.")
            else:
                raise FileNotFoundError(f"Unable to find the instance save file {self.path}`.")


class CachedAsyncChatbotFactory:
    def __init__(self, *, parent=None, hash_prefix: str, **kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)
        
        self.instances: dict[str, AsyncChatbotInstance] = {}
        self.kwargs = kwargs
        self.hash_prefix = hash_prefix
        self.tools: list[PredicateTool] = []
        self.commands: list[PredicateCommand] = []
        self.bot = parent
        assert self.bot

    def register_command(self, command: PredicateCommand):
        """
        Register a command

        commands are inline tools with only one parameter that can be used by all models even without tool
        calling capabilities, their descriptions are dynamically added to the system prompt at runtime
        """

        self.commands.append(command)

    def register_tool(self, tool: PredicateTool):
        """
        Register a tool (function) for tool calling.
        """

        self.tools.append(tool)

    async def __execute_predicate_if_present(self, predicate: AsyncPredicate | None, *args, **kwargs) -> bool:
        if predicate is None:
            return True
        return await predicate(*args, **kwargs)

    async def new_instance(self, identifier: str) -> AsyncChatbotInstance:
        instance = AsyncChatbotInstance(identifier, **self.kwargs, hash_prefix=self.hash_prefix, bot=self.bot)

        # register all the tools for the newly created instance
        for tool in self.tools:
            if not await self.__execute_predicate_if_present(tool.predicate, instance):
                continue
            instance.add_tool(
                name=tool.name,
                func=partial(tool.func, instance),
                parameters=tool.parameters,
                description=tool.description
            )

        # register all the commands for the newly created instance
        for command in self.commands:
            if not await self.__execute_predicate_if_present(command.predicate, instance):
                continue
            instance.add_command(
                name=command.name,
                func=partial(command.func, instance),
                field_name=command.field_name,
                description=command.description
            )

        return instance

    def cache_instance(self, instance: AsyncChatbotInstance):
        self.instances.update({instance.id: instance})

    async def get(self, identifier: str) -> AsyncChatbotInstance | None:
        cached_instance = self.instances.get(identifier)
        if cached_instance:
            return cached_instance
        instance = await self.new_instance(identifier)
        try:
            instance.load(not_found_ok=False)
            # cache the instance
            self.cache_instance(instance)
            return instance
        except FileNotFoundError:
            return None

    async def get_or_create(self, identifier: str) -> AsyncChatbotInstance:
        instance = self.instances.get(identifier)
        if instance is None:
            instance = await self.new_instance(identifier)
            instance.load(not_found_ok=True)
            # cache the instance
            self.cache_instance(instance)
            self.logger.info(f"initiated a conversation with {identifier=}.")
        return instance

    def remove(self, identifier: str):
        self.logger.info(f"removing {identifier}")
        save_path = get_instance_save_path(id=identifier, hash_prefix=self.hash_prefix)
        if os.path.exists(save_path):
            os.remove(save_path)
        if identifier in self.instances.keys():
            del self.instances[identifier]

    def save(self):
        for identifier, conversation in self.instances.items():
            try:
                conversation.save()
            except Exception as e:
                self.logger.exception(f"Failed to save conversation with {identifier=}: {e}")



import os
import sys
import glob
import logging
import importlib.util

from ..utils import PixiPaths

class AddonManager:
    def __init__(self, bot):
        self.bot = bot
        self.addons_dir = str(PixiPaths.addons().absolute())
        self.addons = {}

        os.makedirs(self.addons_dir, exist_ok=True)

    def load_addons(self):
        for addon_path in glob.glob(os.path.join(self.addons_dir, "*_addon")):
            if not os.path.isdir(addon_path):
                continue
            basename = os.path.basename(addon_path)
            addon_name = basename
            if not addon_name.startswith("_"):
                self.load_addon(addon_name, addon_path)

    def load_addon(self, name: str, path: str):
        sys.path.append(os.path.dirname(path))

        addon_file = os.path.join(path, "__init__.py")
        if not os.path.isfile(addon_file):
            logging.warning(f"Addon '{name}' does not have an __init__.py file.")
            return

        self.load_addon_file(name, addon_file)

    def load_addon_file(self, name: str, path: str):
        logging.info(f"Loading addon: {name} from {path}")
        spec = importlib.util.spec_from_file_location(name, path, submodule_search_locations=[])
        if spec and spec.loader:
            module = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(module)
                if hasattr(module, "register"):
                    module.register(self.bot, logging.getLogger(name))
                    self.addons[name] = module
                else:
                    logging.warning(f"Addon '{name}' does not have a valid register function.")
            except Exception as e:
                logging.exception(f"Failed to load addon '{name}'")
        else:
            logging.error(f"Could not load addon '{name}': Invalid module specification.")

    def get_addon(self, name):
        return self.addons.get(name)



import dataclasses
from enum import IntEnum, auto


class FilterType(IntEnum):
    Whitelist = auto()
    Blacklist = auto()
    Allow = auto()
    Deny = auto()


@dataclasses.dataclass(frozen=True)
class IdFilter:
    type: FilterType
    ids: list[str] | None = None

    @classmethod
    def whitelist(cls, ids: list[str]) -> 'IdFilter':
        return IdFilter(
            FilterType.Whitelist, ids=ids
        )

    @classmethod
    def blacklist(cls, ids: list[str]) -> 'IdFilter':
        return IdFilter(
            FilterType.Blacklist, ids=ids
        )

    @classmethod
    def allow(cls) -> 'IdFilter':
        return IdFilter(
            FilterType.Allow
        )

    @classmethod
    def deny(cls) -> 'IdFilter':
        return IdFilter(
            FilterType.Deny
        )

    def is_allowed(self, id: str):
        match self.type:
            case FilterType.Whitelist:
                return id in self.ids if self.ids else True  # yes, even when whitelising, if ids is not defined we allow everything
            case FilterType.Blacklist:
                return id not in self.ids if self.ids else True
            case FilterType.Allow:
                return True
            case FilterType.Deny:
                return False



import os
import tempfile
from typing import Optional

import av

from ..utils import PixiPaths

from .base import MediaCache, CompressedMedia, UnsupportedMediaException


# constants

CACHE_SAMPLERATE = 16000
CACHE_BITRATE = 32000
CACHE_MAX_DURATION = 30


class AudioCache(MediaCache):
    def __init__(self, data_bytes: Optional[bytes] = None, hash_value: Optional[str] = None, strict: bool = False):
        self.strict = strict
        super().__init__(
            str(PixiPaths.cache() / "audio"),
            format="aac",
            mime_type="audio/aac",
            data_bytes=data_bytes,
            hash_value=hash_value
        )

    def compress(self, data_bytes: bytes) -> CompressedMedia:
        with tempfile.NamedTemporaryFile() as tmp_in, tempfile.NamedTemporaryFile(suffix=f".{self.format}") as tmp_out:
            tmp_in.write(data_bytes)
            tmp_in.flush()

            input_file = av.open(tmp_in.name)

            input_stream = input_file.streams.audio[0]
            duration = float(input_stream.duration or 0)
            exceeds_max_duration = duration > CACHE_MAX_DURATION
            if self.strict and exceeds_max_duration:
                raise UnsupportedMediaException(
                    f"Unsupported media type or format: audio should not be longer than {CACHE_MAX_DURATION} secounds."
                )

            output_file = av.open(
                tmp_out.name,
                mode="w",
            )
            output_stream = output_file.add_stream(
                "aac",
                rate=CACHE_SAMPLERATE,
                bit_rate=CACHE_BITRATE,
                layout='mono'
            )

            duration = 0.0
            for frame in input_file.decode(input_stream):
                duration += (frame.duration or 0.0) / frame.rate
                for packet in output_stream.encode(frame):
                    output_file.mux(packet)
                if duration > CACHE_MAX_DURATION:
                    break

            # Flush the stream
            for packet in output_stream.encode(None):
                output_file.mux(packet)

            input_file.close()
            output_file.close()

            tmp_out.seek(0)
            audio_bytes = tmp_out.read()

        return CompressedMedia(
            mime_type=self.mime_type,
            bytes=audio_bytes,
            format=self.format,
            metadata=dict(duration=CACHE_MAX_DURATION, is_cut_off=exceeds_max_duration)
        )



import logging

from dataclasses import dataclass

from .utils import CoroutineQueueExecutor
from .typing import AsyncFunction, Optional, Iterator, Generator, AsyncIterator, AsyncGenerator
from .chatting import ChatMessage


@dataclass
class AsyncCommand:
    name: str
    field_name: str
    function: AsyncFunction
    description: Optional[str] = None

    # TODO: implement callbacks for when we enter the command (after the command name) and when
    # we leave the command (right after the command is completed btu before executing the command)
    enter_callback: Optional[AsyncFunction] = None
    leave_callback: Optional[AsyncFunction] = None

    def get_syntax(self):
        desc = self.description or "no description"
        return f"[{self.name}:<{self.field_name}>]: {desc}"

    async def __call__(self, *args, **kwds):
        return await self.function(*args, **kwds)

    def __str__(self):
        return f"<async-function {self.name}>"


class AsyncCommandManager:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        
        self.commands: dict[str, AsyncCommand] = dict()

    def _add_command(self, command: AsyncCommand):
        assert command is not None
        self.commands.update({command.name.lower(): command})

    def add_command(self, name: str, field_name: str, function: AsyncFunction, description: Optional[str] = None):
        assert name is not None
        assert function is not None
        assert field_name is not None
        self._add_command(AsyncCommand(
            name=name,
            field_name=field_name,
            function=function,
            description=description,
        ))

    def get_prompt(self):
        return "\n".join(["- "+func.get_syntax() for name, func in self.commands.items()])

    async def execute_command(self, command_str: str, reference_message: Optional[ChatMessage] = None):
        command_content = command_str[1:-1]
        seperator_idx = None
        if ":" in command_str:
            seperator_idx = command_content.index(":")

        if seperator_idx:
            command_name = command_content[:seperator_idx].strip()
            command_data = command_content[seperator_idx+1:].strip()
        else:
            command_name = command_content
            command_data = None

        maybe_command_fn = self.commands.get(command_name.lower())
        if maybe_command_fn is None:
            self.logger.error(f"The command `{command_name}` is not implemented.")
            return

        return await maybe_command_fn(reference_message, command_data)

    async def stream_commands(self, stream: Iterator | Generator | AsyncGenerator | AsyncGenerator, reference_message: ChatMessage):
        """
        Consumes commands and runs them automatically
        """
        inside_command = 0  # counts the number of "[" characters minus the number of "]" characters
        command_str = ""
        
        async with CoroutineQueueExecutor() as queue:
            async def process(char):
                nonlocal inside_command
                nonlocal command_str

                result = None

                # the opening of the command
                if char == "[":
                    inside_command += 1

                if inside_command != 0:
                    command_str += char
                else:
                    result = char

                # the closing of the command
                if char == "]":
                    inside_command -= 1

                    # if the command is fully captured
                    if inside_command == 0:
                        # run command without blocking the stream
                        await queue.add_to_queue(self.execute_command(
                            command_str=command_str,
                            reference_message=reference_message
                        ))
                        command_str = ""

                return result

            if isinstance(stream, (Iterator, Generator)):
                for char in stream:
                    result = await process(char)
                    if result:
                        yield result
            elif isinstance(stream, (AsyncIterator, AsyncGenerator)):
                async for char in stream:
                    result = await process(char)
                    if result:
                        yield result
            else:
                raise TypeError(
                    f"expected `stream` to be an Iterator, Generator, AsyncIterator or AsyncGenerator but got `{type(stream)}`!")


import json
import logging
from typing import Optional

from .base import AgentBase
from ..chatting import AsyncChatClient


class RetrievalAgent(AgentBase):
    """
    Retrieves relevant information from context and a query to the agent.
    """

    def __init__(self, context: Optional[list[str]] = None, **client_kwargs):
        super().__init__(**client_kwargs)

        self.context = context or []
        self.system_prompt = "\n".join([
            "## You are a context retrieval agent",
            ""
            "Given a list of entries and a query, you must return any context that is relevent to the query.",
            "Write the response without loosing any data, mention all the details, the less you summerize is the better.",
            ""
            "output a json object with the following keys:",
            " - `relevant`: a list of all information that could possibly be used to answer the query in any way",
            " - `source`: a list of sources where the information was found, if applicable",
            " - `confidence`: a score value between 1 and 10 indicating how confident you are in the information provided",
            ""
            "example output:",
            "```json",
            "{",
            "  \"source\": [\"page_title:Villagers\"]",
            "  \"relevant\": [\"Villagers can be cured from zombie villagers by using a splash potion of weakness and a golden apple.\"],",
            "  \"confidence\": 9",
            "}",
            "```"
        ])
        self.client.set_system(self.system_prompt)

    def to_dict(self) -> dict:
        return super().to_dict() | dict(
            context=self.context
        )

    @classmethod
    def from_dict(cls, data: dict, **client_kwargs) -> 'RetrievalAgent':
        client = AsyncChatClient(**client_kwargs)
        client.load_state_dict(data["client"])
        context = data.get("context", [])
        return cls(
            context = context,
            client = client            
        )

    def add_context(self, context: str):
        logging.debug(f"Adding context: {context}")
        self.context.append(context)

    def format_query(self, query: str) -> str:
        logging.debug(f"Retrieving information for query: {query}")
        prompt = "\n".join([
            "Context:",
            "```json",
            json.dumps(self.context),
            "```",
            ""
            f"Query: \"{query}\"",
        ])
        return prompt


from enum import IntFlag, auto


class PixiFeatures(IntFlag):
    EnableToolCalling = auto()
    EnableToolLogging = auto()
    EnableWikiSearch = auto()
    EnableGIFSearch = auto()
    EnableImageSupport = auto()
    EnableAudioSupport = auto()

    @classmethod
    def all(cls) -> 'PixiFeatures':
        return cls.EnableToolCalling | cls.EnableToolLogging | cls.EnableWikiSearch | cls.EnableGIFSearch | cls.EnableImageSupport | cls.EnableAudioSupport
    
    @classmethod
    def empty(cls) -> 'PixiFeatures':
        return cls(0)
